import time

import pytest
from egse.device import DeviceConnectionError
from egse.device import DeviceTimeoutError
from egse.system import Timer
from egse.tempcontrol.keithley.daq6510_dev import DAQ6510EthernetInterface

SCPI_PORT = 5025
HOSTNAME = "192.168.68.77"  # hostname at home with DHCP


def is_daq6510_available():
    try:
        daq = DAQ6510EthernetInterface(HOSTNAME, SCPI_PORT)
        daq.connect()
    except DeviceTimeoutError:
        return False

    return True


def test_constructor():
    daq = DAQ6510EthernetInterface(HOSTNAME, SCPI_PORT)
    daq.connect()

    response = daq.trans("*IDN?\n")

    assert response == "KEITHLEY INSTRUMENTS,MODEL DAQ6510,04569510,1.7.12b"

    daq.disconnect()


def test_connection(caplog):
    daq = DAQ6510EthernetInterface(HOSTNAME, SCPI_PORT)

    assert not daq.is_connected()

    daq.connect()

    assert daq.is_connected()

    daq.disconnect()

    assert not daq.is_connected()

    daq.disconnect()  # This should do nothing

    daq.connect()

    caplog.clear()

    daq.connect()  # calling this when the connection is open issues a warning

    assert "trying to connect to an already connected socket" in caplog.text

    daq.disconnect()

    assert not daq.is_connected()

    daq.reconnect()
    assert daq.is_connected()

    daq.reconnect()
    assert daq.is_connected()

    daq.disconnect()


def test_context_manager():
    with DAQ6510EthernetInterface(HOSTNAME, SCPI_PORT) as daq:
        assert daq.is_connected()

    assert not daq.is_connected()


def test_incorrect_construction():
    daq = DAQ6510EthernetInterface()
    daq.hostname = "unknown"  # set this explicitly because the local_settings might define he HOSTNAME
    with pytest.raises(DeviceConnectionError, match="DAQ6510: Socket address info error for unknown"):
        daq.connect()

    daq = DAQ6510EthernetInterface()
    daq.hostname = None
    with pytest.raises(ValueError, match="hostname is not initialized"):
        daq.connect()

    daq = DAQ6510EthernetInterface(HOSTNAME)
    daq.port = None

    with pytest.raises(ValueError, match="port number is not initialized"):
        daq.connect()

    daq = DAQ6510EthernetInterface(HOSTNAME, 3000)  # pass an incorrect port number

    with pytest.raises(DeviceConnectionError, match="DAQ6510: Connection refused"):
        daq.connect()


def test_write_read():
    with DAQ6510EthernetInterface(HOSTNAME, SCPI_PORT) as daq:
        daq.write(':SYST:COMM:LAN:CONF "AUTO"')
        daq.write(":SYST:COMM:LAN:CONF?")

        response = daq.read().decode()

        assert response.startswith("auto")


def test_a_scan():
    with DAQ6510EthernetInterface() as daq:
        # Initialize

        daq.write("*RST")  # this also the user-defined buffer "test1"

        for cmd, response in [
            ('TRAC:MAKE "test1", 1000', False),  # create a new buffer
            # settings for channel 1 and 2 of slot 1
            ('SENS:FUNC "TEMP", (@101:102)', False),  # set the function to temperature
            ("SENS:TEMP:TRAN FRTD, (@101)", False),  # set the transducer to 4-wire RTD
            ("SENS:TEMP:RTD:FOUR PT100, (@101)", False),  # set the type of the 4-wire RTD
            ("SENS:TEMP:TRAN RTD, (@102)", False),  # set the transducer to 2-wire RTD
            ("SENS:TEMP:RTD:TWO PT100, (@102)", False),  # set the type of the 2-wire RTD
            ('ROUT:SCAN:BUFF "test1"', False),
            ("ROUT:SCAN:CRE (@101:102)", False),
            ("ROUT:CHAN:OPEN (@101:102)", False),
            ("ROUT:STAT? (@101:102)", True),
            ("ROUT:SCAN:STAR:STIM NONE", False),
            # ("ROUT:SCAN:ADD:SING (@101, 102)", False),  # not sure what this does, not really needed
            ("ROUT:SCAN:COUN:SCAN 1", False),  # not sure if this is needed in this setting
            # ("ROUT:SCAN:INT 1", False),
        ]:
            if response:
                print(f"Sending {cmd}... response=", end="")
                print(daq.trans(cmd))
            else:
                print(f"Sending {cmd}...")
                daq.write(cmd)

        # Read out the channels

        # daq.write('TRAC:CLE "test1"\n')

        for _ in range(10):
            daq.write("INIT:IMM")
            daq.write("*WAI")

            # Reading the data

            # When a trigger mode is running, these READ? commands can not be used.

            # print(daq.trans('READ? "test1", CHAN, TST, READ\n', wait=False), end="")
            # print(daq.trans('READ? "test1", CHAN, TST, READ\n', wait=False), end="")
            # time.sleep(1)
            # print(daq.trans('READ? "test1", CHAN, TST, READ\n', wait=False), end="")
            # print(daq.trans('READ? "test1", CHAN, TST, READ\n', wait=False), end="")

            # Read out the buffer

            response = daq.trans('TRAC:DATA? 1, 2, "test1", CHAN, TST, READ')
            print(f"{response = }")
            ch1, tst1, val1, ch2, tst2, val2 = response.split(",")
            print(f"Channel: {ch1} Time: {tst1} Value: {float(val1):.4f}\t", end="")
            print(f"Channel: {ch2} Time: {tst2} Value: {float(val2):.4f}")
            time.sleep(0.1)


def test_another_scan():
    with DAQ6510EthernetInterface() as daq:
        # Initialize

        n_readings = 50

        daq.write("*RST")  # This also clears the default buffers

        for cmd, response in [
            ('SENS:FUNC "TEMP", (@101,102)', False),
            ("SENS:TEMP:UNIT CELS, (@101,102)", False),
            ("SENS:TEMP:TRAN FRTD, (@101)", False),  # set the transducer to 4-wire RTD
            ("SENS:TEMP:RTD:FOUR PT100, (@101)", False),  # set the type of the 4-wire RTD
            ("SENS:TEMP:TRAN RTD, (@102)", False),  # set the transducer to 2-wire RTD
            ("SENS:TEMP:RTD:TWO PT100, (@102)", False),  # set the type of the 2-wire RTD
            # Set the amount of time that the input signal is measured.
            ("SENS:TEMP:NPLC 0.1, (@101,102)", False),
            ('TRIG:LOAD "Empty"', False),
            ('TRIG:BLOC:BUFF:CLEAR 1, "defbuffer1"', False),
            ('TRIG:BLOC:MDIG 2, "defbuffer1"', False),
            (f"TRIG:BLOC:BRAN:COUN 3, {n_readings}, 2", False),
        ]:
            if response:
                print(f"Sending {cmd}... response=", end="")
                print(daq.trans(cmd))
            else:
                print(f"Sending {cmd}...")
                daq.write(cmd)

        # Read out the channels

        for _ in range(10):
            daq.write("INIT:IMM")
            daq.write("*WAI")

            # Reading the data

            response = daq.trans(f'TRAC:DATA? 1, {n_readings}, "defbuffer1", CHAN, TST, READ')
            print(f"{response = }")
            # ch1, tst1, val1, ch2, tst2, val2 = response.split(",")
            # print(f"Channel: {ch1} Time: {tst1} Value: {float(val1):.4f}\t", end="")
            # print(f"Channel: {ch2} Time: {tst2} Value: {float(val2):.4f}")
            time.sleep(0.1)



import asyncio
import logging
from typing import Any
from typing import Dict
from typing import Optional

from egse.scpi import AsyncSCPIInterface

logger = logging.getLogger(__name__)


class DAQ6510(AsyncSCPIInterface):
    """Keithley DAQ6510 specific implementation."""

    def __init__(self, hostname: str, port: int = 5025, settings: Optional[Dict[str, Any]] = None):
        """Initialize a Keithley DAQ6510 interface.

        Args:
            hostname: Hostname or IP address
            port: TCP port (default 5025 for SCPI)
            settings: Additional device settings
        """
        super().__init__(
            device_name="DAQ6510",
            hostname=hostname,
            port=port,
            settings=settings,
            id_validation="DAQ6510",  # String that must appear in IDN? response
        )

        self._measurement_lock = asyncio.Lock()

    async def initialize(self):
        # Initialize

        await self.write("*RST")  # this also the user-defined buffer "test1"

        for cmd, response in [
            ('TRAC:MAKE "test1", 1000', False),  # create a new buffer
            # settings for channel 1 and 2 of slot 1
            ('SENS:FUNC "TEMP", (@101:102)', False),  # set the function to temperature
            ("SENS:TEMP:TRAN FRTD, (@101)", False),  # set the transducer to 4-wire RTD
            ("SENS:TEMP:RTD:FOUR PT100, (@101)", False),  # set the type of the 4-wire RTD
            ("SENS:TEMP:TRAN RTD, (@102)", False),  # set the transducer to 2-wire RTD
            ("SENS:TEMP:RTD:TWO PT100, (@102)", False),  # set the type of the 2-wire RTD
            ('ROUT:SCAN:BUFF "test1"', False),
            ("ROUT:SCAN:CRE (@101:102)", False),
            ("ROUT:CHAN:OPEN (@101:102)", False),
            ("ROUT:STAT? (@101:102)", True),
            ("ROUT:SCAN:STAR:STIM NONE", False),
            # ("ROUT:SCAN:ADD:SING (@101, 102)", False),  # not sure what this does, not really needed
            ("ROUT:SCAN:COUN:SCAN 1", False),  # not sure if this is needed in this setting
            # ("ROUT:SCAN:INT 1", False),
        ]:
            if response:
                logger.info(f"Sending {cmd}...")
                response = (await self.trans(cmd)).decode().strip()
                logger.info(f"{response = }")
            else:
                logger.info(f"Sending {cmd}...")
                await self.write(cmd)

    async def get_measurement(self, channel: str) -> float:
        """Get a measurement from a specific channel.

        Args:
            channel: Channel to measure (e.g., "101")

        Returns:
            The measured value as a float
        """
        async with self._measurement_lock:
            cmd = "INIT:IMM"
            logger.info(f"Sending {cmd}...")
            await self.write(cmd)
            cmd = "*WAI"
            logger.info(f"Sending {cmd}...")
            await self.write(cmd)

            if channel == "101":
                start_index = end_index = 1
            elif channel == "102":
                start_index = end_index = 2
            else:
                return float("nan")

            response = (
                (await self.trans(f'TRAC:DATA? {start_index}, {end_index}, "test1", CHAN, TST, READ')).decode().strip()
            )

            logger.info(f"{response = }")

            ch, tst, val = response.split(",")

            logger.info(f"Channel: {ch} Time: {tst} Value: {float(val):.4f}")

            return float(val)



# async_gateway.py
import asyncio
import time
from datetime import datetime

import zmq
import zmq.asyncio
from influxdb_client_3 import InfluxDBClient3
from influxdb_client_3 import Point


class AsyncMetricsHub:
    def __init__(
        self,
        influx_client,  # FIXME: should be a plugin
        zmq_endpoint="tcp://*:5555",  # FIXME: should also come from a plugin
        batch_size=500,
        flush_interval=2.0,
    ):
        self.influx_client = influx_client
        self.batch_size = batch_size
        self.flush_interval = flush_interval

        # AsyncIO ZeroMQ context
        self.context = zmq.asyncio.Context()
        self.socket = self.context.socket(zmq.PULL)
        self.socket.bind(zmq_endpoint)

        # Configure socket
        self.socket.setsockopt(zmq.RCVHWM, 10000)

        # AsyncIO queue for data points
        self.data_queue = asyncio.Queue(maxsize=10000)

        # Stats
        self.stats = {"received": 0, "written": 0, "errors": 0, "queue_size": 0}

        print(f"🚀 AsyncIO Gateway listening on {zmq_endpoint}")

    async def start(self):
        """Start the gateway with concurrent tasks"""
        try:
            # Run receiver and batch processor concurrently
            await asyncio.gather(
                self._receive_data(), self._batch_processor(), self._stats_reporter(), return_exceptions=True
            )
        except Exception as e:
            print(f"❌ Gateway error: {e}")
        finally:
            await self._cleanup()

    async def _receive_data(self):
        """Receive data from ZeroMQ and add to queue"""
        print("📡 Starting data receiver...")

        while True:
            try:
                # Receive message (non-blocking with timeout)
                message = await self.socket.recv_json()

                # Process message
                point = await self._process_message(message)
                if point:
                    # Add to queue (non-blocking)
                    try:
                        await asyncio.wait_for(self.data_queue.put(point), timeout=0.1)
                        self.stats["received"] += 1
                    except asyncio.TimeoutError:
                        # Queue full, drop point
                        self.stats["errors"] += 1
                        print("⚠️ Queue full, dropping point")

            except Exception as e:
                print(f"❌ Error receiving data: {e}")
                self.stats["errors"] += 1
                await asyncio.sleep(0.1)

    async def _process_message(self, message):
        """Process incoming message asynchronously"""
        try:
            # Validate message
            if "device_id" not in message or "data" not in message:
                return None

            # Create InfluxDB point
            point = Point("sensor_data")
            point.tag("device_id", message["device_id"])

            # Add timestamp
            if "timestamp" in message:
                point.time(message["timestamp"])

            # Add sensor data
            sensor_data = message["data"]
            for key, value in sensor_data.items():
                if isinstance(value, (int, float)):
                    point.field(key, value)
                else:
                    point.tag(key, str(value))

            return point

        except Exception as e:
            print(f"❌ Error processing message: {e}")
            return None

    async def _batch_processor(self):
        """Batch processor using asyncio"""
        print("📦 Starting batch processor...")

        batch = []
        last_flush = time.time()

        while True:
            try:
                # Try to get data points for batch
                timeout = max(0.1, self.flush_interval - (time.time() - last_flush))

                try:
                    # Get point with timeout
                    point = await asyncio.wait_for(self.data_queue.get(), timeout=timeout)
                    batch.append(point)

                except asyncio.TimeoutError:
                    # No data received, check if we should flush anyway
                    pass

                # Update queue size stat
                self.stats["queue_size"] = self.data_queue.qsize()

                # Check flush conditions
                should_flush = len(batch) >= self.batch_size or (
                    batch and time.time() - last_flush >= self.flush_interval
                )

                if should_flush and batch:
                    await self._flush_batch(batch)
                    batch.clear()
                    last_flush = time.time()

            except Exception as e:
                print(f"❌ Batch processor error: {e}")
                await asyncio.sleep(1)

    async def _flush_batch(self, batch):
        """Flush batch to InfluxDB asynchronously"""
        if not batch:
            return

        try:
            start_time = time.time()

            # Run InfluxDB write in thread pool to avoid blocking
            loop = asyncio.get_event_loop()
            await loop.run_in_executor(None, self.influx_client.write, batch)

            write_time = time.time() - start_time
            self.stats["written"] += len(batch)

            points_per_second = len(batch) / write_time if write_time > 0 else 0
            print(f"✅ Wrote {len(batch)} points ({points_per_second:.0f} pts/s)")

        except Exception as e:
            print(f"❌ Failed to write batch: {e}")
            self.stats["errors"] += len(batch)

    async def _stats_reporter(self):
        """Periodic stats reporting"""
        while True:
            await asyncio.sleep(10)  # Report every 10 seconds
            print(
                f"📊 Stats: Received={self.stats['received']}, "
                f"Written={self.stats['written']}, "
                f"Queue={self.stats['queue_size']}, "
                f"Errors={self.stats['errors']}"
            )

    async def _cleanup(self):
        """Cleanup resources"""
        print("🧹 Cleaning up...")

        # Flush remaining data
        remaining_batch = []
        while not self.data_queue.empty():
            try:
                point = await asyncio.wait_for(self.data_queue.get(), timeout=0.1)
                remaining_batch.append(point)
            except asyncio.TimeoutError:
                break

        if remaining_batch:
            await self._flush_batch(remaining_batch)

        self.socket.close()
        self.context.term()
        print("✅ Cleanup complete")


# Device process (also async)
class AsyncDeviceDataSender:
    def __init__(self, gateway_endpoint="tcp://localhost:5555", device_id="device_001"):
        self.device_id = device_id
        self.context = zmq.asyncio.Context()
        self.socket = self.context.socket(zmq.PUSH)
        self.socket.connect(gateway_endpoint)
        self.socket.setsockopt(zmq.LINGER, 0)
        self.socket.setsockopt(zmq.SNDHWM, 1000)

    async def send_data_point(self, sensor_data):
        """Send data point asynchronously"""
        message = {"device_id": self.device_id, "timestamp": datetime.utcnow().isoformat(), "data": sensor_data}

        try:
            await self.socket.send_json(message, flags=zmq.NOBLOCK)
            return True
        except zmq.Again:
            # Queue full
            return False

    async def close(self):
        self.socket.close()
        self.context.term()


# Usage examples
async def device_data_loop():
    """Example device data collection loop"""
    sender = AsyncDeviceDataSender(device_id="sensor_001")

    try:
        while True:
            # Your sensor reading logic here
            sensor_data = {
                "temperature": 23.5 + (time.time() % 10),  # Mock data
                "humidity": 45.0 + (time.time() % 5),
                "pressure": 1013.25,
            }

            success = await sender.send_data_point(sensor_data)
            if not success:
                print("⚠️ Failed to send data point")

            await asyncio.sleep(0.1)  # 10Hz

    except KeyboardInterrupt:
        print("🛑 Stopping device loop...")
    finally:
        await sender.close()


async def run_gateway():
    """Run the gateway service"""
    # Initialize InfluxDB client
    influx_client = InfluxDBClient3(host="your-influxdb-host", token="your-token", database="your-database")

    gateway = AsyncInfluxGateway(influx_client=influx_client, batch_size=250, flush_interval=1.0)

    await gateway.start()


# Main execution
if __name__ == "__main__":
    try:
        # Run gateway
        asyncio.run(run_gateway())

        # Or run device simulator
        # asyncio.run(device_data_loop())

    except KeyboardInterrupt:
        print("👋 Goodbye!")



import logging
import time

from egse.signal import FileBasedSignaling
from egse.signal import create_signal_command_file

logging.basicConfig(level=logging.INFO)

logger = logging.getLogger("egse.test_signal")


def test_reregister():
    service_id = "SIGNAL-TEST"

    signaling = FileBasedSignaling(service_id)
    signaling.start_monitoring()

    create_signal_command_file(signaling.signal_dir, service_id, {"action": "reregister", "params": {"x": 42, "y": 23}})

    running = True

    def stop():
        nonlocal running

        logger.warning("Request to terminate the test.")

        running = False

    signaling.register_handler("stop", stop)

    count = 10

    while running:
        signaling.process_pending_commands()
        time.sleep(0.1)
        if count == 0:
            # This will be called after 10 seconds and end the test
            create_signal_command_file(signaling.signal_dir, service_id, {"action": "stop"})
        count -= 1

    signaling.stop()



from __future__ import annotations

__all__ = [
    "create_empty_file",
    "create_text_file",
    "is_process_not_running",
    "setup_conf_data",
    "setup_data_storage_layout",
    "teardown_data_storage_layout",
    "is_service_registry_running",
]

import os
import textwrap
from pathlib import Path
from typing import List

from egse.env import get_site_id
from egse.env import set_conf_data_location
from egse.env import set_data_storage_location
from egse.env import set_log_file_location
from egse.process import is_process_running
from egse.registry.client import RegistryClient


def is_process_not_running(items: List):
    """Check if a process is not running currently."""
    return not is_process_running(items)


def is_service_registry_running() -> bool:
    client = RegistryClient()
    return client.health_check()


def setup_conf_data(tmp_data_dir: Path):
    site_id = get_site_id()
    data_root = tmp_data_dir / site_id / "conf"
    data_root.mkdir(parents=True, exist_ok=True)

    create_text_file(
        data_root / f"SETUP_{site_id}_00000_240123_120000.yaml",
        textwrap.dedent(
            f"""\
            # This is the 'Zero' Setup for {site_id}.

            Setup:
                site_id: {site_id}

                history:
                    0: Initial zero Setup for {site_id}
            """
        ),
    )

    create_text_file(
        data_root / f"SETUP_{site_id}_00028_240123_120028.yaml",
        textwrap.dedent(
            f"""\
            # This is Setup nr 28 for {site_id}.

            Setup:
                site_id: {site_id}

                history:
                    0: Initial zero Setup for {site_id}
                    28: I just jumped straight to twenty eight
            """
        ),
    )


def teardown_conf_data(data_dir: Path): ...


def teardown_data_storage_layout(data_dir: Path): ...


def setup_data_storage_layout(tmp_data_dir: Path) -> Path:
    """
    Create a standard layout for the data storage as expected by the CGSE. The path is created from the
    `tmp_data_dir`. The site_id is derived from the environment module using: `get_site_id()`.

    The layout with site_id = "LAB23":

        tmp_data_dir/
        └───data/
            └── LAB23
                ├── conf
                │   └── data
                ├── daily
                │   └── 20250118
                ├── log
                └── obs

    Returns:
        The path to the data folder including the site_id. In the above case,
        that is: `{tmp_data_dir}/data/LAB23`.

    """
    data_root = tmp_data_dir / get_site_id()
    data_root.mkdir(parents=True)

    tmp_dir = data_root / "daily"
    tmp_dir.mkdir()

    tmp_dir = data_root / "conf"
    tmp_dir.mkdir()

    tmp_dir = data_root / "conf" / "data"
    tmp_dir.mkdir()

    tmp_dir = data_root / "obs"
    tmp_dir.mkdir()

    tmp_dir = data_root / "log"
    tmp_dir.mkdir()

    set_data_storage_location(str(data_root))
    set_conf_data_location(str(data_root / "conf"))
    set_log_file_location(str(data_root / "log"))

    return data_root


def create_empty_file(filename: str | Path, create_folder: bool = False):
    """
    A function and context manager to create an empty file with the given
    filename. When used as a function, the file needs to be removed explicitly
    with a call to `filename.unlink()` or `os.unlink(filename)`.

    This function can be called as a context manager in which case the file will
    be removed when the context ends.

    Returns:
        The filename as a Path.
    """

    class _ContextManager:
        def __init__(self, filename: str | Path, create_folder: bool):
            self.filename = Path(filename)

            if self.filename.exists():
                raise FileExistsError(f"The empty file you wanted to create already exists: {filename}")

            if create_folder and not self.filename.parent.exists():
                self.filename.parent.mkdir(parents=True)

            with self.filename.open(mode="w"):
                pass

        def __enter__(self):
            return self.filename

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.filename.unlink()

    return _ContextManager(filename, create_folder)


def create_text_file(filename: str | Path, content: str, create_folder: bool = False):
    """
    A function and context manager to create a text file with the given string
    as content. When used as a function, the file needs to be removed explicitly
    with a call to `filename.unlink()` or `os.unlink(filename)`.

    This function can be called as a context manager in which case the file will
    be removed when the context ends.

    >> with create_text_file("samples.txt", "A,B,C\n1,2,3\n4,5,6\n"):
    ..     # do something with the file or its content

    Returns:
        The filename as a Path.
    """

    class _ContextManager:
        def __init__(self, filename: str | Path, create_folder: bool):
            self.filename = Path(filename)

            if self.filename.exists():
                raise FileExistsError(f"The text file you wanted to create already exists: {filename}")

            if create_folder and not self.filename.parent.exists():
                self.filename.parent.mkdir(parents=True)

            with filename.open(mode="w") as fd:
                fd.write(content)

        def __enter__(self):
            return self.filename

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.filename.unlink()

    return _ContextManager(filename, create_folder)


# Test the helper functions


def main():
    print(f"cwd = {os.getcwd()}")

    fn = Path("xxx.txt")

    with create_empty_file(fn):
        assert fn.exists()
    assert not fn.exists()

    create_empty_file(fn)
    assert fn.exists()
    fn.unlink()
    assert not fn.exists()

    # Test the create_a_text_file() helper function

    with create_text_file(
        fn,
        textwrap.dedent(
            """\
        A,B,C,D
        1,2,3,4
        5,6,7,8
        """
        ),
    ) as filename:
        assert fn.exists()
        assert filename == fn

        print(fn.read_text())

    assert not fn.exists()

    fn = Path("data/xxx.txt")

    with create_empty_file(fn, create_folder=True):
        assert fn.exists()

    assert not fn.exists()


if __name__ == "__main__":
    main()



import pytest

from egse.registry.service import ZMQMicroservice
from egse.system import get_host_ip


@pytest.mark.asyncio
async def test_zmq_microservice_initialization():
    print()

    service = ZMQMicroservice("vanilla", "plain")

    assert service.service_name == "vanilla"
    assert service.service_type == "plain"
    assert service.host_ip == get_host_ip()

    print(f"{service.rep_port = }")

    assert service.rep_port != 0

    await service._cleanup()


# Registration to the service registry will time out in 5s
@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_zmq_microservices_registration(caplog):
    print()

    service = ZMQMicroservice("vanilla", "plain")

    caplog.clear()
    await service.start()
    assert "Failed to register with the service registry" in caplog.text



"""
Test program for the registry server and clients.

Start the registry server as follows. We use these non-standard ports to be completely independent and not interfere
with running core services:

    $ uv run py -m egse.registry.server start --log-level DEBUG --req-port 15556 --pub-port 15557 --hb-port=15558 \
     --db-path test_service_registry.db

Start the clients as follows:

    $ uv run py libs/cgse-core/tests/script_test_registry_client_server.py --use-test-ports

    $ uv run py libs/cgse-core/tests/script_test_registry_client_server.py --use-asyncio --use-test-ports

Notes:
  - the `--use-test-ports` option is the same as using: `--req-port 15556 --pub-port 15557 --hb-port=15558`,
    but this only works for the test clients
  - make sure to use the `--db-path` option when starting the server, or you will use the same SQLite database
    as the core service.
  - you should be able to start several clients, both async and no-async.
  - if you want to get status of the test server, its sufficient to use:

    $  uv run py -m egse.registry.server status --req-port 15556

  - you can stop the test server with CTRL-C or with the command:

    $  uv run py -m egse.registry.server stop --req-port 15556

"""

import asyncio
import logging
import sys
import time

import typer

from egse.registry import DEFAULT_RS_HB_PORT
from egse.registry import DEFAULT_RS_PUB_PORT
from egse.registry import DEFAULT_RS_REQ_PORT
from egse.registry.client import AsyncRegistryClient
from egse.registry.client import RegistryClient


# Constants for testing
TEST_REQ_PORT = 15556
TEST_PUB_PORT = 15557
TEST_HB_PORT = 15558

# Wait timeout for the server to start (seconds)
SERVER_STARTUP_TIMEOUT = 5


def test_proper_termination_of_tasks_sync(
    req_port: int = DEFAULT_RS_REQ_PORT,
    pub_port: int = DEFAULT_RS_PUB_PORT,
    hb_port: int = DEFAULT_RS_HB_PORT,
    host: str = "localhost",
):
    client = RegistryClient(
        registry_req_endpoint=f"tcp://{host}:{req_port}",
        registry_sub_endpoint=f"tcp://{host}:{pub_port}",
        registry_hb_endpoint=f"tcp://{host}:{hb_port}",
        request_timeout=5000,  # 5 second timeout
    )
    client.connect()

    if not client.health_check():
        raise RuntimeError("Client failed to connect to server")

    service_id = client.register(
        name="sync-context-test-service",
        host="localhost",
        port=8080,
        service_type="context-test",
        metadata={"msg": "Hello, World!"},
        ttl=10,
    )

    client.start_heartbeat()

    response = client.get_service(service_id)

    print(response)

    assert response is not None
    assert response["name"] == "sync-context-test-service"
    assert response["metadata"]["msg"] == "Hello, World!"

    print("Sleeping for 10s to let some heartbeats come through...")
    time.sleep(50.0)

    client.stop_heartbeat()
    client.deregister(service_id)
    client.disconnect()


async def test_proper_termination_of_tasks_async(
    req_port: int = DEFAULT_RS_REQ_PORT,
    pub_port: int = DEFAULT_RS_PUB_PORT,
    hb_port: int = DEFAULT_RS_HB_PORT,
    host: str = "localhost",
):
    client = AsyncRegistryClient(
        registry_req_endpoint=f"tcp://{host}:{req_port}",
        registry_sub_endpoint=f"tcp://{host}:{pub_port}",
        registry_hb_endpoint=f"tcp://{host}:{hb_port}",
        request_timeout=5000,  # 5 second timeout
    )
    client.connect()

    # Verify client can connect to server
    health = await client.health_check()
    if not health:
        raise RuntimeError("Client failed to connect to server")

    service_id = await client.register(
        name="async-context-test-service",
        host="localhost",
        port=8080,
        service_type="context-test",
        metadata={"msg": "Hello, World!"},
        ttl=10,
    )

    await client.start_heartbeat()

    response = await client.get_service(service_id)

    print(response)

    assert response is not None
    assert response["name"] == "async-context-test-service"
    assert response["metadata"]["msg"] == "Hello, World!"

    print("Sleeping for 10s to let some heartbeats come through...")
    await asyncio.sleep(10.0)

    await client.stop_heartbeat()
    await client.deregister(service_id)
    client.disconnect()


app = typer.Typer()


@app.command()
def main(
    req_port: int = DEFAULT_RS_REQ_PORT,
    pub_port: int = DEFAULT_RS_PUB_PORT,
    hb_port: int = DEFAULT_RS_HB_PORT,
    host: str = "localhost",
    use_asyncio: bool = False,
    use_test_ports: bool = False,
):
    if use_test_ports:
        req_port = TEST_REQ_PORT
        pub_port = TEST_PUB_PORT
        hb_port = TEST_HB_PORT

    if use_asyncio:
        asyncio.run(test_proper_termination_of_tasks_async(req_port, pub_port, hb_port, host))
    else:
        test_proper_termination_of_tasks_sync(req_port, pub_port, hb_port, host)


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.DEBUG,
        format="[%(asctime)s] %(threadName)-12s %(levelname)-8s %(name)-12s %(lineno)5d:%(module)-20s %(message)s",
    )

    sys.exit(app())



import contextlib
import logging
import re
import sys
import time

import pytest

from egse.confman import is_configuration_manager_active
from egse.dummy import DummyProxy
from egse.dummy import is_dummy_cs_active
from egse.dummy import is_dummy_dev_active
from egse.process import SubProcess
from egse.proxy import Proxy
from egse.response import Failure
from egse.system import Timer
from egse.system import type_name
from egse.system import waiting_for

logger = logging.getLogger("egse.tests")


def is_service_running(services: list) -> bool:
    is_active = False

    try:
        if "cm" in services:
            if is_configuration_manager_active():
                is_active = True
            else:
                logger.warning("Configuration Manager is not running.")
                is_active = False
    except Exception as exc:
        logger.error(f"Caught {type_name(exc)}: {exc}")
        is_active = False

    return is_active


# Skip entire module if service not running
if not is_service_running(["cm"]):
    pytest.skip("Core service 'cm' not available", allow_module_level=True)


@contextlib.contextmanager
def dummy_service():
    dummy_dev = SubProcess("Dummy Device", [sys.executable, "-m", "egse.dummy"], ["start-dev"])
    dummy_dev.execute()

    with Timer(name="Dummy Dev startup timer"):
        waiting_for(is_dummy_dev_active, timeout=5.0)

    dummy_cs = SubProcess("Dummy CS", [sys.executable, "-m", "egse.dummy"], ["start-cs"])
    dummy_cs.execute()

    processes = [dummy_dev, dummy_cs]

    # It takes ~1.5s to startup on my MacBook Pro M2, why is this such a long time?
    with Timer(name="Dummy CS startup timer"):
        waiting_for(is_dummy_cs_active, timeout=5.0)

    yield processes

    dummy_cs_stop = SubProcess("Dummy CS", [sys.executable, "-m", "egse.dummy"], ["stop-cs"])
    dummy_cs_stop.execute()

    # It takes ~1.5s to startup on my MacBook Pro M2, why is this such a long time?
    with Timer(name="Dummy CS shutdown timer"):
        waiting_for(lambda: not is_dummy_cs_active(), timeout=5.0)

    dummy_dev_stop = SubProcess("Dummy Device", [sys.executable, "-m", "egse.dummy"], ["stop-dev"])
    dummy_dev_stop.execute()

    with Timer(name="Dummy Dev shutdown timer"):
        waiting_for(lambda: not is_dummy_dev_active(), timeout=5.0)


def is_valid_ip_address_format(string):
    pattern = r"^\d{0,3}\.\d{0,3}\.\d{0,3}\.\d{0,3}$"
    return bool(re.match(pattern, string))


def test_proxy_without_cs():
    proxy = DummyProxy()

    assert proxy.is_cs_connected() is False
    assert proxy.has_commands() is False
    assert proxy.get_commands() == []

    with pytest.raises(NotImplementedError):
        assert proxy.info()


def test_device_commands():
    with dummy_service():
        with DummyProxy() as dummy:
            with Timer("info", log_level=logging.WARNING, precision=6):
                assert dummy.info().startswith("Dummy Device")

            with Timer("get_value", log_level=logging.WARNING, precision=6):
                assert 0.0 <= dummy.get_value() < 1.0

            assert dummy.division(144, 12) == 12
            response = dummy.division(33, 0)
            assert isinstance(response, Failure)
            assert response.successful is False
            assert response.message == "Executing division failed: : division by zero"
            assert isinstance(response.cause, ZeroDivisionError)


def test_protocol_commands():
    with dummy_service():
        with DummyProxy() as dummy:
            with Timer("ping", log_level=logging.WARNING, precision=6):
                assert dummy.ping() is True

            assert dummy.get_commanding_port() == 4443
            assert dummy.get_service_port() == 4444
            assert dummy.get_monitoring_port() == 4445

            assert dummy.get_endpoint() == "tcp://localhost:4443"

            assert dummy.has_commands() is True

            for cmd in "info", "get_value", "handle_event":
                assert cmd in dummy.get_commands()

            assert is_valid_ip_address_format(dummy.get_ip_address())

            assert isinstance(dummy.get_service_proxy(), Proxy)
            assert dummy.is_cs_connected() is True


def test_dummy_service():
    print()

    with dummy_service() as procs:
        for proc in procs:
            print(f"{proc.is_running()=}")

        with DummyProxy() as dummy:
            assert dummy.ping() is True

    for proc in procs:
        print(f"{proc.is_running()=}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module defines some plotting functions for Reference Frames.

@author: Pierre Royer
"""

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # needed for 3d projection
import numpy as np
import egse
from egse.coordinates.referenceFrame import ReferenceFrame
from egse.coordinates.point import Point, Points


def plot_reference_frame(frame, master=None, figname=None, **kwargs):
    """Plot a Reference Frame.

    Args:
        frame  : egse.coordinates.referenceFrame.ReferenceFrame
        master : master ReferenceFrame (optional)
        figname: string. Name of matplotlib figure (can be pre-existing)
        kwargs : passed to matplotlib.axes._subplots.Axes3DSubplot.quiver

    Returns:
        matplotlib.axes._subplots.Axes3DSubplot displaying the reference frame.

        The three unit vectors are shown with the following colors ('RGB'):
            x: Red
            y: Green
            z: Blue

    .. note::
        Use ax.set_xlim3d(min,max) to properly set the ranges of the display

    """
    if master is None:
        tmpmaster = ReferenceFrame.createMaster()
    else:
        tmpmaster = master.__copy__()

    f0 = frame.getOrigin()
    fx = frame.getAxis("x", name="fx")
    fy = frame.getAxis("y", name="fy")
    fz = frame.getAxis("z", name="fz")
    f0m = f0.expressIn(tmpmaster)[:3]
    fxm = fx.expressIn(tmpmaster)[:3]
    fym = fy.expressIn(tmpmaster)[:3]
    fzm = fz.expressIn(tmpmaster)[:3]
    del tmpmaster

    # Origin of the X,Y and Z vectors (x = the 'x' coordinates of the origin of all 3 vectors)
    # Every vector independently (--> plot in diff. colors)
    x, y, z = np.array([f0m[0]]), np.array([f0m[1]]), np.array([f0m[2]])

    # Orientation of the X,Y and Z vectors
    vecxx, vecyx, veczx = np.array([fxm[0] - f0m[0]]), np.array([fym[0] - f0m[0]]), np.array([fzm[0] - f0m[0]])
    vecxy, vecyy, veczy = np.array([fxm[1] - f0m[1]]), np.array([fym[1] - f0m[1]]), np.array([fzm[1] - f0m[1]])
    vecxz, vecyz, veczz = np.array([fxm[2] - f0m[2]]), np.array([fym[2] - f0m[2]]), np.array([fzm[2] - f0m[2]])

    kwargs.setdefault("length", 1)
    kwargs.setdefault("normalize", True)
    # kwargs.setdefault('figsize', (10,10))

    fig = plt.figure(figname, figsize=plt.figaspect(1.0))
    ax = fig.add_subplot(projection="3d")
    ax.quiver(x, y, z, vecxx, vecxy, vecxz, color="r", **kwargs)
    ax.quiver(x, y, z, vecyx, vecyy, vecyz, color="g", **kwargs)
    ax.quiver(x, y, z, veczx, veczy, veczz, color="b", **kwargs)
    # ax.axis('equal')

    return ax


def plot_points(points, master=None, figname=None, **kwargs):
    """Plot the Points object.

    Args:
        points : either a (egse.coordinate.point.)Points object or a list of Point objects
        master : master ReferenceFrame (optional)
        figname: string. Name of matplotlib figure (can be pre-existing)
        kwargs : passed to matplotlib.axes._subplots.Axes3DSubplot.scatter

    Returns:
        matplotlib.axes._subplots.Axes3DSubplot displaying the reference frame.

    .. note::
        Use ax.set_xlim3d(min,max) to properly set the ranges of the display.

    """
    if master is None:
        tmpmaster = ReferenceFrame.createMaster()
    else:
        tmpmaster = master.__copy__()
    #
    if isinstance(points, list):
        allpoints = Points(points, ref=tmpmaster)
    elif isinstance(points, Points) or isinstance(points, egse.coordinates.point.Points):
        allpoints = points
    else:
        raise ValueError("If the input is a list, all items in it must be Point objects")
    #
    del tmpmaster
    #
    coordinates = allpoints.coordinates
    xs = coordinates[0, :]
    ys = coordinates[1, :]
    zs = coordinates[2, :]
    #
    kwargs.setdefault("s", 50)
    kwargs.setdefault("marker", "o")
    kwargs.setdefault("color", "k")
    #
    fig = plt.figure(figname)
    ax = fig.add_subplot(projection="3d")
    ax.scatter(xs, ys, zs, **kwargs)

    return ax


def plot_vectors(points, master=None, figname=None, fromorigin=True, **kwargs):
    """Plot the Points object.

    Args:
        points : either a (egse.coordinate.point.)Points object or a list of Point objects
        master : master ReferenceFrame (optional)
        figname: string. Name of matplotlib figure (can be pre-existing)
        fromorigin: bool
                    if True, all vectors are displayed starting from the origin
                    if False, all vectors go towards the origin
        kwargs : passed to matplotlib.axes._subplots.Axes3DSubplot.scatter

    Returns:
        matplotlib.axes._subplots.Axes3DSubplot displaying the reference frame.

    .. note::
        Use ax.set_xlim3d(min,max) to properly set the ranges of the display.

    """

    if master is None:
        tmpmaster = ReferenceFrame.createMaster()
    else:
        tmpmaster = master.__copy__()
    #
    if isinstance(points, list):
        allpoints = Points(points, ref=tmpmaster)
    elif isinstance(points, Points) or isinstance(points, egse.coordinates.point.Points):
        allpoints = points
    else:
        raise ValueError("If the input is a list, all items in it must be Point objects")
    #
    del tmpmaster
    #

    # SET DEFAULTS
    kwargs.setdefault("color", "k")
    #

    # PREPARE VECTOR COORDINATES
    coordinates = allpoints.coordinates
    xs = coordinates[0, :]
    ys = coordinates[1, :]
    zs = coordinates[2, :]

    # Origin of the X,Y and Z vectors
    # ==> x = the 'x' coordinates of the origin of all vectors)
    # ==> [x,y,z] = the origin of points.ref
    x, y, z = points.ref.getOrigin().coordinates[:3]
    x = np.ones_like(xs) * x
    y = np.ones_like(xs) * y
    z = np.ones_like(xs) * z

    # PLOT

    fig = plt.figure(figname)
    ax = fig.gca(projection="3d")

    if fromorigin:
        ax.quiver(x, y, z, xs - x, ys - y, zs - z, **kwargs)

    elif not fromorigin:
        ax.quiver(xs, ys, zs, x - xs, y - ys, z - zs, **kwargs)

    else:
        print("Parameter 'fromorigin' must be True or False")
        print("Setting it to True by default")
        ax.quiver(x, y, z, xs - x, ys - y, zs - z, **kwargs)

    return ax



"""
The Process Manager Control Server is the service that handles all Common-EGSE
processes.  The list of processes to handle, is taken from the Configuration
Manager (from the setup).

The Process Manager Control Server is implemented as a standard control server.
"""

import logging
import multiprocessing

import rich
import sys
import typer
import zmq

from egse.confman import ConfigurationManagerProxy
from egse.control import ControlServer
from egse.listener import EVENT_ID
from egse.logger import remote_logging
from egse.process import SubProcess
from egse.procman import ProcessManagerProxy
from egse.procman.procman_protocol import ProcessManagerProtocol
from egse.registry.client import RegistryClient
from egse.services import ServiceProxy
from egse.settings import Settings
from egse.storage import store_housekeeping_information
from egse.zmq_ser import get_port_number

# Use explicit name here otherwise the logger will probably be called __main__

logger = logging.getLogger("egse.procman")

CTRL_SETTINGS = Settings.load("Process Manager Control Server")


class ProcessManagerControlServer(ControlServer):
    def __init__(self):
        super().__init__()

        multiprocessing.current_process().name = app_name

        self.logger = logger
        self.service_name = app_name
        self.service_type = CTRL_SETTINGS.SERVICE_TYPE

        self.device_protocol = ProcessManagerProtocol(self)

        self.logger.debug(f"Binding ZeroMQ socket to {self.device_protocol.get_bind_address()}")

        self.register_as_listener(
            proxy=ConfigurationManagerProxy,
            listener={"name": "Process Manager CS", "proxy": ProcessManagerProxy, "event_id": EVENT_ID.SETUP},
        )
        self.device_protocol.bind(self.dev_ctrl_cmd_sock)

        self.poller.register(self.dev_ctrl_cmd_sock, zmq.POLLIN)

        self.register_service(service_type=CTRL_SETTINGS.SERVICE_TYPE)

        self.set_hk_delay(10.0)

        self.logger.info(f"PM housekeeping saved every {self.hk_delay / 1000:.1f} seconds.")

    def get_communication_protocol(self):
        return "tcp"

    def get_commanding_port(self):
        return get_port_number(self.dev_ctrl_cmd_sock) or 0

    def get_service_port(self):
        return get_port_number(self.dev_ctrl_service_sock) or 0

    def get_monitoring_port(self):
        return get_port_number(self.dev_ctrl_mon_sock) or 0

    def get_storage_mnemonic(self):
        try:
            return CTRL_SETTINGS.STORAGE_MNEMONIC
        except AttributeError:
            return "PM"

    def is_storage_manager_active(self):
        from egse.storage import is_storage_manager_active

        return is_storage_manager_active()

    def store_housekeeping_information(self, data):
        origin = self.get_storage_mnemonic()
        store_housekeeping_information(origin, data)

    def register_to_storage_manager(self):
        from egse.storage import register_to_storage_manager
        from egse.storage.persistence import TYPES

        register_to_storage_manager(
            origin=self.get_storage_mnemonic(),
            persistence_class=TYPES["CSV"],
            prep={
                "column_names": list(self.device_protocol.get_housekeeping().keys()),
                "mode": "a",
            },
        )

    def unregister_from_storage_manager(self):
        from egse.storage import unregister_from_storage_manager

        unregister_from_storage_manager(origin=self.get_storage_mnemonic())

    def after_serve(self):
        from egse.confman import ConfigurationManagerProxy

        self.unregister_as_listener(proxy=ConfigurationManagerProxy, listener={"name": "Process Manager CS"})

        self.deregister_service()


app_name = "pm_cs"
app = typer.Typer(name=app_name)


@app.command()
def start():
    """Starts the Process Manager (pm_cs).

    The pm_cs is normally started automatically on egse-server boot.
    """

    multiprocessing.current_process().name = "pm_cs"

    with remote_logging():
        try:
            control_server = ProcessManagerControlServer()
            control_server.serve()
        except KeyboardInterrupt:
            print("Shutdown requested...exiting")
        except SystemExit as exit_code:
            print(f"System Exit with code {exit_code}.")
            sys.exit(exit_code.code)
        except Exception:
            import traceback

            traceback.print_exc(file=sys.stdout)

    return 0


@app.command()
def start_bg():
    """Starts the Process Manager Control Server in the background."""

    proc = SubProcess("pm_cs", ["pm_cs", "start"])
    proc.execute()


@app.command()
def stop():
    """Sends a 'quit_server' command to the Process Manager."""

    with RegistryClient() as reg:
        service = reg.discover_service(CTRL_SETTINGS.SERVICE_TYPE)
        # rich.print("service = ", service)

        if service:
            with ServiceProxy(hostname=service["host"], port=service["metadata"]["service_port"]) as proxy:
                proxy.quit_server()
        else:
            rich.print("[red]ERROR: Couldn't connect to 'pm_cs', process probably not running.")


@app.command()
def status():
    """Prints the status of the control server."""

    import rich
    from egse.procman import get_status

    rich.print(get_status(), end="")


if __name__ == "__main__":
    sys.exit(app())



"""
This module manages files that have a counter in their filename.
"""

__all__ = [
    "counter_exists",
    "counter_filename",
    "get_next_counter",
    "new_counter",
]

from pathlib import Path

from egse.config import find_files
from egse.log import logger


def counter_filename(location: Path, filename: Path | str) -> Path:
    """
    Creates an absolute filename to be used as a counter file. A counter file usually has a 'count' extension
    but that is not enforced by this module. The location can be a relative path, even '.' or '..' are accepted.

    Args:
        location: the location of the counter file.
        filename: the name of the counter file, use the '.count' extension.

    Returns:
         An absolute filename.

    Note:
        If the file doesn't exist, it is NOT created.

    """
    return (location / filename).resolve()


def counter_exists(filename: Path) -> bool:
    """
    Returns True if the given file exists, False otherwise.

    Args:
        filename: path of the counter file

    Returns:
        True if the given filename exists, False otherwise.

    Raises:
        OSError: might be raised by Path.

    Note:
        No checking is done if the file is indeed a counter file, i.e. if it contains the correct content.
          So, this function basically only checks if the given Path exists and if it is a regular file.

    """
    return filename.exists() if filename.is_file() else False


def new_counter(filename: Path, pattern: str) -> int:
    """
    Create a counter based on the files that already exist for the given pattern.

    Args:
        filename: the name of the counter file
        pattern: a pattern to match the filenames

    Returns:
        The next counter value as an integer.
    """
    if not counter_exists(filename):
        location = filename.parent
        counter = determine_counter_from_dir_list(location, pattern)
        _write_counter(counter, filename)
    else:
        counter = get_next_counter(filename)

    return counter


def _write_counter(counter: int, filename: Path):
    """
    Overwrites the given counter in the given file. The file contains nothing else then the counter.
    If the file didn't exist before, it will be created.

    Args:
        counter: the counter to save
        filename: the file to which the counter shall be saved
    """
    with filename.open("w") as fd:
        fd.write(f"{counter:d}")


def _read_counter(filename: Path) -> int:
    """
    Reads a counter from the given file. The file shall only contain the counter which must
    be an integer on the first line of the file. If the given file doesn't exist, 0 is returned.

    Args:
        filename: the full path of the file containing the counter

    Returns:
        The counter that is read from the file or 0 if file doesn't exist.
    """
    try:
        with filename.open("r") as fd:
            counter = fd.read().strip()
    except FileNotFoundError:
        counter = 0
    return int(counter or 0)


def get_next_counter(filename: Path) -> int:
    """
    Read the counter from a dedicated file, add one and save the counter back to the file.

    Args:
        filename: full pathname of the file that contains the required counter

    Returns:
        The value of the next counter, 1 if no previous files were found or if an error occurred.

    Note:
        This will create the counter file if it doesn't exist.
    """

    counter = _read_counter(filename)
    counter += 1
    _write_counter(counter, filename)

    return counter


def determine_counter_from_dir_list(location: Path | str, pattern: str, index: int = -1) -> int:
    """
    Determine counter for a new file at the given location and with the given pattern.
    The next counter is determined from the sorted list of files that match the given pattern.

    Args:
        location: Location where the file should be stored.
        pattern: Pattern for the filename.
        index: the location of the counter in the filename after it is split on '_' [default=-1]

    Returns:
        The value of the next counter, 1 if no previous files were found or if an error occurred.
    """

    files = sorted(find_files(pattern=pattern, root=location))

    # No filenames found showing the given pattern -> start counting at 1

    if len(files) == 0:
        return 1

    last_file = files[-1]

    parts = last_file.name.split("_")

    try:
        # Observation files have the following pattern:
        #  <test ID>_<lab ID>_<setup ID>_<storage mnemonic>_<day YYYYmmdd>_<time HHMMSS>[_<counter>]
        #
        # Daily files:
        #  <day>_<site ID>_<storage mnemonic>[_<counter>]
        #
        # Any file:
        #  the counter is assumed to be the last part before the file extension and is preceded by an underscore.
        #  <anything here>_counter>.<extension>

        counter = int(parts[index].split(".")[0]) + 1
        logger.debug(f"{counter = }")
        return counter

    except ValueError:
        logger.warning("ValueError", exc_info=True)
        return 1



"""
This module defines the device classes to be used to connect to and control the Hexapod JORAN from
Symétrie.

"""

import logging
import math
import time

from egse.hexapod.symetrie.alpha import AlphaPlusControllerInterface
from egse.hexapod.symetrie.dynalpha import AlphaPlusTelnetInterface, decode_validation_error
from egse.mixin import DynamicCommandMixin
from egse.proxy import DynamicProxy
from egse.settings import Settings
from egse.system import Timer
from egse.system import wait_until
from egse.zmq_ser import connect_address

logger = logging.getLogger(__name__)

JORAN_SETTINGS = Settings.load("JORAN Controller")
CTRL_SETTINGS = Settings.load("Hexapod JORAN Control Server")
DEVICE_SETTINGS = Settings.load(filename="joran.yaml")


class JoranInterface(AlphaPlusControllerInterface):
    """
    Interface definition for the JoranController, the JoranProxy, and the JoranSimulator.
    """


class JoranController(JoranInterface, DynamicCommandMixin):
    def __init__(self):
        self.hostname = {JORAN_SETTINGS.IP}
        self.port = {JORAN_SETTINGS.PORT}
        self.transport = self.hexapod = AlphaPlusTelnetInterface(self.hostname, self.port)

        super.__init__()

    def is_simulator(self):
        return False

    def is_connected(self):
        return self.hexapod.is_connected()

    def connect(self):
        self.hexapod.connect()

    def disconnect(self):
        self.hexapod.disconnect()

    def reconnect(self):
        if self.is_connected():
            self.disconnect()
        self.connect()

    def reset(self, wait=True):
        raise NotImplementedError

    # def sequence(self):
    #     raise NotImplementedError

    def set_virtual_homing(self, tx, ty, tz, rx, ry, rz):
        raise NotImplementedError

    def get_debug_info(self):
        raise NotImplementedError

    def jog(self, axis: int, inc: float) -> int:
        raise NotImplementedError

    def get_temperature(self):
        raise NotImplementedError

    def get_limits_state(self):
        raise NotImplementedError

    def machine_limit_enable(self, state):
        raise NotImplementedError

    def user_limit_set(self, *par):
        raise NotImplementedError

    def set_default(self):
        raise NotImplementedError


class JoranSimulator(JoranInterface):
    """
    HexapodSimulator simulates the Symétrie Hexapod JORAN. The class is heavily based on the
    ReferenceFrames in the `egse.coordinates` package.

    The simulator implements the same methods as the HexapodController class which acts on the
    real hardware controller in either simulation mode or with a real Hexapod JORAN connected.

    Therefore, the HexapodSimulator can be used instead of the Hexapod class in test harnesses
    and when the hardware is not available.

    This class simulates all the movements and status of the Hexapod.
    """

    def __init__(self):
        super().__init__()

        # Keep a record if the homing() command has been executed.

        self.homing_done = False
        self.control_loop = False
        self._virtual_homing = False
        self._virtual_homing_position = None

    def is_simulator(self):
        return True

    def connect(self):
        pass

    def reconnect(self):
        pass

    def disconnect(self):
        # TODO:
        #   Should I keep state in this class to check if it has been disconnected?
        #
        # TODO:
        #   What happens when I re-connect to this Simulator? Shall it be in Homing position or
        #   do I have to keep state via a persistence mechanism?
        pass

    def is_connected(self):
        return True

    def clear_error(self):
        return 0

    def homing(self):
        self.goto_zero_position()
        self.homing_done = True
        self._virtual_homing = False
        self._virtual_homing_position = None
        return 0

    def is_homing_done(self):
        return self.homing_done

    def activate_control_loop(self):
        self.control_loop = True
        return self.control_loop

    def deactivate_control_loop(self):
        self.control_loop = False
        return self.control_loop

    pass


class JoranProxy(DynamicProxy, JoranInterface):
    """The JoranProxy class is used to connect to the control server and send commands to the
    Hexapod JORAN remotely."""

    def __init__(
        self,
        protocol=CTRL_SETTINGS.PROTOCOL,
        hostname=CTRL_SETTINGS.HOSTNAME,
        port=CTRL_SETTINGS.COMMANDING_PORT,
    ):
        """
        Args:
            protocol: the transport protocol [default is taken from settings file]
            hostname: location of the control server (IP address) [default is taken from settings
            file]
            port: TCP port on which the control server is listening for commands [default is
            taken from settings file]
        """
        super().__init__(connect_address(protocol, hostname, port))


if __name__ == "__main__":
    from rich import print as rp

    joran = JoranController()
    joran.connect()

    with Timer("JoranController"):
        rp(joran.info())
        rp(joran.is_homing_done())
        rp(joran.is_in_position())
        rp(joran.activate_control_loop())
        rp(joran.get_general_state())
        rp(joran.get_actuator_state())
        rp(joran.deactivate_control_loop())
        rp(joran.get_general_state())
        rp(joran.get_actuator_state())
        rp(joran.stop())
        rp(joran.get_limits_value(0))
        rp(joran.get_limits_value(1))
        rp(joran.check_absolute_movement(1, 1, 1, 1, 1, 1))
        rp(joran.check_absolute_movement(51, 51, 51, 1, 1, 1))
        rp(joran.get_speed())
        rp(joran.set_speed(2.0, 1.0))
        time.sleep(0.5)  # if we do not sleep, the get_speed() will get the old values
        speed = joran.get_speed()

        if not math.isclose(speed["vt"], 2.0):
            rp(f"[red]{speed['vt']} != 2.0[/red]")
        if not math.isclose(speed["vr"], 1.0):
            rp(f"[red]{speed['vr']} != 1.0[/red]")

        rp(joran.get_actuator_length())

        # rp(joran.machine_limit_enable(0))
        # rp(joran.machine_limit_enable(1))
        # rp(joran.get_limits_state())
        rp(joran.get_coordinates_systems())
        rp(
            joran.configure_coordinates_systems(
                0.033000,
                -0.238000,
                230.205000,
                0.003282,
                0.005671,
                0.013930,
                0.000000,
                0.000000,
                0.000000,
                0.000000,
                0.000000,
                0.000000,
            )
        )
        rp(joran.get_coordinates_systems())
        rp(joran.get_machine_positions())
        rp(joran.get_user_positions())
        rp(
            joran.configure_coordinates_systems(
                0.000000,
                0.000000,
                0.000000,
                0.000000,
                0.000000,
                0.000000,
                0.000000,
                0.000000,
                0.000000,
                0.000000,
                0.000000,
                0.000000,
            )
        )
        rp(joran.validate_position(1, 0, 0, 0, 0, 0, 0, 0))
        rp(joran.validate_position(1, 0, 0, 0, 50, 0, 0, 0))

        rp(joran.goto_zero_position())
        rp(joran.is_in_position())
        if wait_until(joran.is_in_position, interval=1, timeout=300):
            rp("[red]Task joran.is_in_position() timed out after 30s.[/red]")
        rp(joran.is_in_position())

        rp(joran.get_machine_positions())
        rp(joran.get_user_positions())

        rp(joran.move_absolute(0, 0, 12, 0, 0, 10))

        rp(joran.is_in_position())
        if wait_until(joran.is_in_position, interval=1, timeout=300):
            rp("[red]Task joran.is_in_position() timed out after 30s.[/red]")
        rp(joran.is_in_position())

        rp(joran.get_machine_positions())
        rp(joran.get_user_positions())

        rp(joran.move_absolute(0, 0, 0, 0, 0, 0))

        rp(joran.is_in_position())
        if wait_until(joran.is_in_position, interval=1, timeout=300):
            rp("[red]Task joran.is_in_position() timed out after 30s.[/red]")
        rp(joran.is_in_position())

        rp(joran.get_machine_positions())
        rp(joran.get_user_positions())

        # joran.reset()
        joran.disconnect()

        rp(0, decode_validation_error(0))
        rp(11, decode_validation_error(11))
        rp(8, decode_validation_error(8))
        rp(24, decode_validation_error(24))



import asyncio
import json
import logging
import multiprocessing
import sys
import time
from typing import Any
from typing import Callable

import typer
import zmq
import zmq.asyncio

from egse.log import logger
from egse.logger import remote_logging
from egse.notifyhub import DEFAULT_COLLECTOR_PORT
from egse.notifyhub import DEFAULT_PUBLISHER_PORT
from egse.notifyhub import DEFAULT_REQUESTS_PORT
from egse.notifyhub import STATS_INTERVAL
from egse.notifyhub.client import AsyncNotificationHubClient
from egse.registry import MessageType
from egse.registry.client import AsyncRegistryClient
from egse.system import TyperAsyncCommand
from egse.system import get_host_ip
from .event import NotificationEvent

RECV_REQUEST_TIMEOUT = 1.0

app = typer.Typer(name="notify_hub")


class AsyncNotificationHub:
    def __init__(self):
        self.server_id = "notification-hub-1"

        self.context: zmq.asyncio.Context = zmq.asyncio.Context()

        # Receive events from services (PULL socket for load balancing)
        self.collector_socket: zmq.asyncio.Socket = self.context.socket(zmq.PULL)

        # Publish events to subscribers (PUB socket for fan-out)
        self.publisher_socket: zmq.asyncio.Socket = self.context.socket(zmq.PUB)

        # Health check socket (ROUTER - can handle multiple clients)
        self.requests_socket: zmq.asyncio.Socket = self.context.socket(zmq.ROUTER)

        # Register notification hub to the service registry
        self.registry_client = AsyncRegistryClient(request_timeout=200)  # FIXME: ms -> s and CONSTANT
        self.service_id = None
        self.service_name = "Notification Hub"
        self.service_type = "notification-hub"

        self.stats = {
            "events_received": 0,
            "events_published": 0,
            "last_message_time": 0.0,
            "start_time": time.time(),
        }

        self.running = False
        self._shutdown_event = asyncio.Event()

        # Tasks
        self._tasks = set()

        self.logger = logging.getLogger("egse.notifyhub")

    async def start(self):
        """Start the notification hub. This will start the event collector and
        publisher, the health check, and the stats reporter as asyncio Tasks.
        """
        multiprocessing.current_process().name = "notifyhub"

        self.running = True
        self.logger.info("Starting Async Notification Hub...")

        self.collector_socket.bind(f"tcp://*:{DEFAULT_COLLECTOR_PORT}")
        self.publisher_socket.bind(f"tcp://*:{DEFAULT_PUBLISHER_PORT}")
        self.requests_socket.bind(f"tcp://*:{DEFAULT_REQUESTS_PORT}")

        self._tasks = [
            asyncio.create_task(self._event_collector()),
            asyncio.create_task(self._stats_reporter()),
            asyncio.create_task(self._handle_requests()),
        ]

        await self._register_service()

        await self._shutdown_event.wait()

        await self._deregister_service()

        await self._shutdown()

    async def _shutdown(self):
        self.running = False
        self.logger.info("Async Notification Hub shutdown requested...")

        # Wait for tasks to gracefully complete (with timeout)
        if self._tasks:
            done, pending = await asyncio.wait(self._tasks, timeout=2.0)
            for task in pending:
                task.cancel()

        self.collector_socket.close()
        self.publisher_socket.close()
        self.requests_socket.close()

        self.context.term()

        self.logger.info("Async Notification Hub shutdown complete")

    async def _register_service(self):
        self.logger.info("Registering service...")

        self.registry_client.connect()

        self.service_id = await self.registry_client.register(
            name=self.service_name,
            host=get_host_ip(),
            port=DEFAULT_REQUESTS_PORT,
            service_type=self.service_type,
            metadata={"pub_port": DEFAULT_PUBLISHER_PORT, "collector_port": DEFAULT_COLLECTOR_PORT},
        )

        if not self.service_id:
            self.logger.error("Failed to register with the service registry")
        else:
            await self.registry_client.start_heartbeat()

    async def _deregister_service(self):
        self.logger.info("De-registering service...")

        if self.service_id:
            await self.registry_client.stop_heartbeat()
            await self.registry_client.deregister()

        self.registry_client.disconnect()

    async def _event_collector(self):
        """Main event collection loop"""
        while self.running:
            try:
                # Receive event from any service (non-blocking with timeout)
                message_bytes = await asyncio.wait_for(self.collector_socket.recv(), timeout=0.1)

                message = json.loads(message_bytes.decode())

                event = NotificationEvent(
                    event_type=message["event_type"],
                    source_service=message["source_service"],
                    data=message["data"],
                    timestamp=message.get("timestamp"),
                    correlation_id=message.get("correlation_id"),
                )

                self.logger.info(f"Received: {event.event_type} from {event.source_service}")

                self.stats["events_received"] += 1
                self.stats["last_message_time"] = time.time()

                await self._publish_event(event)

            except asyncio.TimeoutError:
                continue
            except asyncio.CancelledError:
                self.logger.warning("Event collector cancelled.")
                self.running = False
            except Exception as exc:
                self.logger.error(f"Error in event collector: {exc}", exc_info=True)
                # Why waiting 1s here?
                # - Is this to prevent overloading the logger when there is a serious problem
                # - Will this improve when we get more experience with the NotifyHub?
                await asyncio.sleep(1.0)

    async def _publish_event(self, event: NotificationEvent):
        """Publish event to all subscribers"""
        try:
            await self.publisher_socket.send_multipart(
                [
                    event.event_type.encode(),  # Topic for filtering
                    json.dumps(event.as_dict()).encode(),  # Event data
                ]
            )

            self.stats["events_published"] += 1
            self.logger.info(f"Published: {event.event_type}")

        except Exception as exc:
            self.logger.error(f"Error publishing event: {exc}")

    async def _stats_reporter(self):
        """Periodically report statistics"""
        while self.running:
            try:
                await asyncio.wait_for(self._shutdown_event.wait(), timeout=STATS_INTERVAL)
                # shutdown was requested if we get here
                # self.logger.info("Shutdown request caught in stats reporter.")
                break
            except asyncio.TimeoutError:
                # normal case, 30s timeout before next report
                # FIXME: Shouldn't this also be reported on the PUB channel?
                self.logger.info(f"Stats: {self.stats}")
            except asyncio.CancelledError:
                self.logger.warning("Stats reporter cancelled.")
                self.running = False

    async def _handle_requests(self):
        """Simple health check endpoint simulation"""
        self.logger.info("Started request handler task")

        while self.running:
            try:
                try:
                    message_parts = await asyncio.wait_for(
                        self.requests_socket.recv_multipart(), timeout=RECV_REQUEST_TIMEOUT
                    )
                except asyncio.TimeoutError:
                    # self.logger.debug("waiting for command request...")
                    continue

                if len(message_parts) >= 3:
                    client_id = message_parts[0]
                    message_type = MessageType(message_parts[1])
                    message_data = message_parts[2]

                    self.logger.info(f"{client_id = }, {message_type = }, {message_data = }")
                    response = await self._process_request(message_data)

                    await self._send_response(client_id, message_type, response)

            except zmq.ZMQError as exc:
                self.logger.error(f"ZMQ error: {exc}", exc_info=True)
            except Exception as exc:
                self.logger.error(f"Error handling request: {exc}", exc_info=True)
            except asyncio.CancelledError:
                self.logger.warning("Request handling cancelled.")
                self.running = False

    async def _process_request(self, msg_data: bytes):
        """
        Process a client request and generate a response.

        Args:
            msg_data: the actual JSON with the request

        """
        try:
            request = json.loads(msg_data.decode())
            self.logger.info(f"Received request: {request}")

        except json.JSONDecodeError as exc:
            self.logger.error(f"Invalid JSON received: {exc}")
            return {"success": False, "error": "Invalid JSON format"}

        action = request.get("action")
        if not action:
            return {"success": False, "error": "Missing required field: action"}

        handlers: dict[str, Callable] = {
            "health": self._handle_health,
            "terminate": self._handle_terminate,
        }

        handler = handlers.get(action)
        if not handler:
            return {"success": False, "error": f"Unknown action: {action}"}

        return await handler(request)

    async def _send_response(self, client_id: bytes, msg_type: MessageType, response: dict[str, Any]):
        """
        If the client expects a reply, send the response.

        Args:
            client_id: the client identification, part 1 of the multipart message
            msg_type: the type of the message, e.g. if reply is required
            response: a dictionary with the status and response

        """
        if msg_type == MessageType.REQUEST_WITH_REPLY:
            await self.requests_socket.send_multipart(
                [client_id, MessageType.RESPONSE.value, json.dumps(response).encode()]
            )

    async def _handle_health(self, request: dict[str, Any]) -> dict[str, Any]:
        """Handle a health check request."""

        self.logger.info(f"Handle health request: {request}")

        return {"success": True, "status": "ok", "timestamp": int(time.time())}

    async def _handle_terminate(self, request: dict[str, Any]) -> dict[str, Any]:
        """Handle a termination request."""

        self.logger.info(f"Handle termination request: {request}")

        await self.stop()

        return {
            "success": True,
            "status": "terminating",
            "timestamp": time.time(),
        }

    async def _generate_health_response(self, request):
        """Generate health response data"""
        health_data = {
            "status": "healthy" if self.running else "unhealthy",
            "uptime": time.time() - self.stats["start_time"],
            "events_received": self.stats["events_received"],
            "events_published": self.stats["events_published"],
            "last_message_time": self.stats["last_message_time"],
            "timestamp": time.time(),
            "request_id": request.get("request_id"),
            "server_id": self.server_id,
        }

        return health_data

    async def stop(self):
        """Signal the server to stop."""
        self._shutdown_event.set()


@app.command(cls=TyperAsyncCommand)
# Usage
async def start():
    with remote_logging():
        hub = AsyncNotificationHub()
        await hub.start()


@app.command(cls=TyperAsyncCommand)
async def stop():
    with AsyncNotificationHubClient() as client:
        await client.terminate_notification_hub()


if __name__ == "__main__":
    try:
        rc = app()
    except zmq.ZMQError as exc:
        if "Address already in use" in str(exc):
            logger.error(f"The Service Registry server is already running: {exc}")
        else:
            logger.error("Couldn't start service registry server", exc_info=True)
        rc = -1

    except KeyboardInterrupt:
        logging.info("KeyboardInterrupt received for NotifyHub, terminating...")
        rc = -1

    sys.exit(rc)



__all__ = [
    "DEFAULT_SIGNAL_DIR",
    "FileBasedSignaling",
    "create_signal_command_file",
]

import json
import os
import queue
import tempfile
import threading
import time
from pathlib import Path

from egse.log import logger
from egse.system import format_datetime

DEFAULT_SIGNAL_DIR = "/tmp/cgse_signals"


class FileBasedSignaling:
    """
    Thread-safe file-based signaling system for inter-process communication.

    Monitors a directory for JSON command files and queues them for processing
    in the main thread. Commands are automatically removed after being read.

    Usage:
        signaling = FileBasedSignaling("my_service_id")
        signaling.start_monitoring()
        signaling.register_handler('reload', my_reload_function)

        # In your main loop:
        signaling.process_pending_commands()

        # Send commands by creating JSON files:
        echo '{"action": "reload", "params": {}}' > /tmp/cgse_signals/my_service_id_cmd.json

    Thread Safety:
        - File monitoring runs in a background daemon thread
        - Commands are queued and processed with process_pending_commands()
          which should be called from the main thread, main loop.
        - All command handlers execute in the main thread context

    Args:
        service_id: the identifier of your service that shall execute the command.
        signal_dir (str): Directory path to monitor for command files.
            Created if it doesn't exist. Defaults to "/tmp/cgse_signals"

    To trigger actions, create the files in `/tmp/cgse_signals`:

      $ echo "{'action': 'reregister'}" > /tmp/cgse_signals/cm_cs_reregister.json

    """

    def __init__(self, service_id: str, signal_dir=DEFAULT_SIGNAL_DIR):
        logger.info(f"Set up file-based signaling for {service_id} in {signal_dir!s}")

        self.signal_dir: Path = Path(signal_dir)
        self.signal_dir.mkdir(exist_ok=True)
        self.service_id = service_id
        self.running = True
        self.thread = None  # the monitoring thread

        # Queue to pass commands from monitoring thread to main thread
        self.command_queue = queue.Queue()
        self.command_handlers = {}

        # Register some default handlers
        self.register_handler("reregister", self.handle_reregister)
        self.register_handler("reload", self.handle_reload)

    def register_handler(self, action, handler_func):
        """Register a handler function for a specific action."""

        overwrite = True if action in self.command_handlers else False

        self.command_handlers[action] = handler_func
        logger.info(f"Registered handler for action: {action}{' (overwritten)' if overwrite else ''}")

    def start_monitoring(self):
        """
        Start a monitoring thread for signal command file. The thread will put any commands that match
        the service_id on the command queue to be processed by the main loop.
        """
        self.thread = threading.Thread(target=self._monitor_loop, daemon=True)
        self.thread.start()
        return self.thread

    def _monitor_loop(self):
        while self.running:
            for filepath in self.signal_dir.glob(f"{self.service_id}*.json"):
                if filepath.is_file():
                    try:
                        with open(filepath, "r") as fd:
                            command = json.load(fd)
                        # File will only be deleted when json could load it
                        filepath.unlink(missing_ok=True)

                        # Put command in queue for main thread to process
                        self.command_queue.put(
                            {"command": command, "source": str(filepath), "timestamp": format_datetime()},
                        )
                    except Exception as exc:
                        logger.error(f"Error reading command file {filepath!s}: {exc}")

            time.sleep(0.1)

    # Remember that this function needs to be called from the main thread

    def process_pending_commands(self):
        """
        Process the commands that are on the command queue. The registered command handler that
        matches the action will be called with any keyword arguments that are specified in the
        `param` key.

        Call this from your main thread to process signaling commands.
        """

        processed_count = 0

        # Process all pending commands
        while not self.command_queue.empty():
            try:
                command_data = self.command_queue.get_nowait()
                command = command_data["command"]
                action = command.get("action")
                params: dict = command.get("params", {})

                if action in self.command_handlers:
                    logger.info(f"🎯 Processing command in main thread: {action} with {params}")
                    try:
                        self.command_handlers[action](**params)
                        processed_count += 1
                    except Exception as exc:
                        logger.error(f"Error in handler for {action}: {exc}")
                else:
                    logger.warning(f"No handler registered for action: {action}")

                self.command_queue.task_done()

            except queue.Empty:
                break
            except Exception as exc:
                logger.error(f"Error processing command: {exc}")

        return processed_count

    def handle_reregister(self, **params):
        """Default dummy handler for reregistration signals."""
        logger.warning(f"🔄 Re-registering service with params: {params} -> register your own function")

    def handle_reload(self, **params):
        """Default dummy handler for reloading signals."""
        logger.warning(f"⚙️ Reloading config with params: {params} -> register your own function")

    def stop(self):
        """Terminate the monitoring loop thread."""
        self.running = False
        if self.thread.is_alive():
            self.thread.join(timeout=1)


def create_signal_command_file(signal_dir: Path, service_id: str, command: dict) -> None:
    """
    Atomically create a signal command file for a specific service.

    Creates a JSON command file using atomic file operations to prevent race
    conditions. The file is first written to a temporary file, then atomically
    renamed to the final destination.

    If the signal_dir doesn't exist, it will be created by this function. The
    signal_dir will not be removed afterward.

    Args:
        signal_dir (Path): Directory where the signal file will be created
        service_id (str): Unique identifier for the target service
        command (dict): Command dictionary containing at least an 'action' key.
                       Will be serialized as JSON with 2-space indentation.

    Returns:
        None

    Raises:
        OSError: If the signal directory is not writable or disk is full
        json.JSONEncodeError: If the command cannot be serialized to JSON
        Exception: Re-raises any other exception after cleaning up temp file

    Example:
        signal_dir = Path("/tmp/myapp_signals")
        create_signal_command_file(
            signal_dir,
            "user-service-8001",
            {"action": "reregister", "params": {"force": True}}
        )
        # Creates: /tmp/myapp_signals/user-service-8001_reregister.json

    Note:
        The final filename format is: {service_id}_{action}.json
        If no 'action' key exists in command, defaults to 'cmd'.
    """
    signal_dir.mkdir(exist_ok=True)
    temp_fd, temp_path = tempfile.mkstemp(dir=signal_dir, suffix=".tmp")

    try:
        filepath = signal_dir / f"{service_id}_{command.get('action', 'cmd')}.json"

        with os.fdopen(temp_fd, "w") as fd:
            json.dump(command, fd, indent=2)

        Path(temp_path).rename(filepath)

    except Exception as exc:
        Path(temp_path).unlink(missing_ok=True)
        raise exc



import sys
import time
from typing import List

import pytest
import rich

from egse.confman import ConfigurationManagerProxy
from egse.confman import is_configuration_manager_active
from egse.dummy import is_dummy_cs_active
from egse.process import SubProcess
from egse.system import Timer
from egse.system import waiting_for


def test_is_cm_cs_is_active():
    assert is_configuration_manager_active() in (False, True)  # Should not raise an exception


@pytest.mark.skipif(not is_configuration_manager_active(), reason="core service cm_cs not running")
def test_list_setups():
    rich.print()

    with ConfigurationManagerProxy() as cm:
        setups = cm.list_setups()

    assert isinstance(setups, List)

    # FIXME: This check is dependent on the current environment that was set up to run the core services

    assert setups[0] == ("00000", "VACUUM_LAB", "Initial zero Setup for VACUUM_LAB", "no sut_id")


@pytest.mark.skipif(not is_configuration_manager_active(), reason="core service cm_cs not running")
def test_load_setup():
    with ConfigurationManagerProxy() as cm:
        setup = cm.load_setup(setup_id=0)
        assert setup.get_id() == "00000"

        # load_setup(..) does change the Setup that is loaded on the cm_cs

        setup = cm.load_setup(setup_id=1)
        assert setup.get_id() == "00001"

        setup = cm.get_setup()
        assert setup.get_id() == "00001"


@pytest.mark.skipif(not is_configuration_manager_active(), reason="core service cm_cs not running")
def test_get_setup():
    with ConfigurationManagerProxy() as cm:
        setup = cm.load_setup(setup_id=0)
        assert setup.get_id() == "00000"

        # get_setup(..) doesn't change the Setup that is loaded on the cm_cs

        setup = cm.get_setup(setup_id=1)
        assert setup.get_id() == "00001"

        setup = cm.get_setup()
        assert setup.get_id() == "00000"


@pytest.mark.skipif(not is_configuration_manager_active(), reason="core service cm_cs not running")
def test_listeners():
    dummy_dev = SubProcess("Dummy Device", [sys.executable, "-m", "egse.dummy", "start-dev"])
    dummy_dev.execute()

    try:
        with ConfigurationManagerProxy() as cm:
            assert "Dummy CS" not in cm.get_listener_names()

            dummy_cs = SubProcess("Dummy CS", [sys.executable, "-m", "egse.dummy", "start-cs"])
            dummy_cs.execute()

            # It takes ~1.5s to startup on my MacBook Pro M2, why is this such a long time?
            with Timer(name="Dummy CS startup timer"):
                waiting_for(is_dummy_cs_active, timeout=5.0)

            assert "Dummy CS" in cm.get_listener_names()

            cm.load_setup(setup_id=1)

            dummy_cs_stop = SubProcess("Dummy CS", [sys.executable, "-m", "egse.dummy"], ["stop-cs"])
            dummy_cs_stop.execute()

            # It takes ~1.5s to startup on my MacBook Pro M2, why is this such a long time?
            with Timer(name="Dummy CS shutdown timer"):
                waiting_for(lambda: not is_dummy_cs_active(), timeout=5.0)

    finally:
        dummy_dev_stop = SubProcess("Dummy Device", [sys.executable, "-m", "egse.dummy"], ["stop-dev"])
        dummy_dev_stop.execute()

        time.sleep(0.5)  # give the processes the time to shut down

        while dummy_dev.is_running():
            time.sleep(1.0)



import datetime
import os
import random
import time

import pandas as pd
from influxdb_client_3 import Point
from rich import print

from egse.metrics import get_metrics_repo
from egse.plugins.metrics.influxdb import InfluxDBRepository
from egse.system import Timer
from egse.system import format_datetime
from egse.system import str_to_datetime


def test_influxdb_access():
    token = os.environ.get("INFLUXDB3_AUTH_TOKEN")

    influxdb = get_metrics_repo("influxdb", {"host": "http://localhost:8181", "database": "ARIEL", "token": token})
    influxdb.connect()

    result = influxdb.query("SELECT * FROM cm ORDER BY TIME DESC LIMIT 20", mode="pandas")
    print(result)
    assert isinstance(result, pd.DataFrame)

    result = influxdb.get_values_last_hours("cm", "cm_setup_id", hours=24, mode="pandas")
    print(f"Values from 'cm_setup_id': \n{result}")

    result = influxdb.get_values_in_range("cm", "cm_site_id", "2025-06-26T08:10:00Z", "2025-06-26T08:15:00Z", mode="")
    print(f"Values from 'cm_site_id': \n{result}")

    result = influxdb.get_table_names()
    print(f"Tables in ARIEL: {result}")

    result = influxdb.get_column_names("cm")
    print(f"Columns in cm: {result}")

    result = influxdb.get_column_names("storagecontrolserver")
    print(f"Columns in storagecontrolserver: {result}")

    result = influxdb.get_column_names("unit_test")
    print(f"Columns in unit_test: {result}")

    influxdb.close()


def test_influxdb_write():
    token = os.environ.get("INFLUXDB3_AUTH_TOKEN")

    influxdb = get_metrics_repo("influxdb", {"host": "http://localhost:8181", "database": "ARIEL", "token": token})

    print()

    with influxdb:
        points = []
        measurements = ["unit_test", "ariel", "localhost", "random"]
        for count in range(500):
            points.append(
                Point.measurement(random.choice(measurements))
                .tag("site_id", "KUL")
                .tag("origin", "UNITTEST")
                .field("count", count)
                .field("random", random.randint(0, 100))
                .time(datetime.datetime.now(tz=datetime.timezone.utc))
            )
            # with Timer("influxdb.write", precision=3):
            #     influxdb.write(points[count])
            # print('.', end="", flush=True)

        print()
        with Timer("influxdb.writes", precision=3):
            influxdb.write(points)

        names = influxdb.get_table_names()
        print(f"{names = }")
        assert "unit_test" in names

        names = influxdb.get_column_names("unit_test")
        assert "count" in names

        result = influxdb.get_values_last_hours("unit_test", "count", hours=1, mode="pandas")
        print(f"Values from 'count': {result}")

        result = influxdb.query("SELECT * FROM unit_test ORDER BY TIME DESC LIMIT 20", mode="pandas")
        print(result)


def test_speed():
    import time

    class DiagnosticInfluxDBRepository(InfluxDBRepository):
        def write_points(self, points):
            start_time = time.time()

            try:
                super().write(points)

            except Exception as exc:
                print(f"Write error: {exc}")
                raise

            finally:
                duration = time.time() - start_time
                print(f"Wrote {len(points)} points in {duration:.3f}s")

                if duration > 0.1:  # Warn if > 100ms
                    print(f"⚠️  Slow write detected: {duration:.3f}s for {len(points)} points")

    token = os.environ.get("INFLUXDB3_AUTH_TOKEN")
    influxdb = DiagnosticInfluxDBRepository(host="http://localhost:8181", database="ARIEL", token=token)
    influxdb.connect()

    points = []
    for count in range(10):
        points.append(
            {
                "measurement": "unit_test",
                "tags": {"site_id": "KUL", "origin": "UNITTEST"},
                "fields": {
                    "count": count,
                    "random": random.randint(0, 100),
                },
                "time": str_to_datetime(format_datetime()),
            }
        )

    influxdb.write_points(points)

    influxdb.close()


def test_connection_speed():
    import requests

    start = time.time()
    response = requests.get("http://localhost:8181/health")
    duration = time.time() - start

    print(f"Health check took {duration:.3f}s")
    if duration > 0.1:
        print("⚠️  Network latency detected")



"""
This module defines the abstract class for any Control Server and some convenience functions.
"""

import abc
import asyncio
import datetime
import logging
import pickle
import textwrap
from functools import partial
from typing import Callable
from typing import Coroutine
from typing import Type
from typing import Union

import zmq
import zmq.asyncio

from egse.decorators import retry
from egse.decorators import retry_with_exponential_backoff
from egse.listener import EVENT_ID
from egse.listener import Event
from egse.listener import Listeners
from egse.system import SignalCatcher

try:
    # This function is only available when the cgse-core package is installed
    from egse.logger import close_all_zmq_handlers
except ImportError:

    def close_all_zmq_handlers():  # noqa
        pass


from egse.process import ProcessStatus
from egse.settings import Settings
from egse.system import get_average_execution_time
from egse.system import get_average_execution_times
from egse.system import get_full_classname
from egse.system import get_host_ip
from egse.system import save_average_execution_time

_LOGGER = logging.getLogger(__name__)

PROCESS_SETTINGS = Settings.load("PROCESS")


async def is_control_server_active(endpoint: str = None, timeout: float = 0.5) -> bool:
    """Checks if the Control Server is running.

    This function sends a *Ping* message to the Control Server and expects a *Pong* answer back within the timeout
    period.

    Args:
        endpoint (str): Endpoint to connect to, i.e. <protocol>://<address>:<port>
        timeout (float): Timeout when waiting for a reply [s, default=0.5]

    Returns: True if the Control Server is running and replied with the expected answer; False otherwise.
    """

    if endpoint is None:
        raise ValueError(
            "endpoint argument not provided, please provide a string with this format: '<protocol>://<address>:<port>'"
        )

    ctx = zmq.asyncio.Context.instance()

    return_code = False

    try:
        socket = ctx.socket(zmq.REQ)
        socket.connect(endpoint)
        data = pickle.dumps("Ping")
        await socket.send(data)

        # Use asyncio.wait_for instead of zmq.select
        try:
            data = await asyncio.wait_for(socket.recv(), timeout=timeout)
            response = pickle.loads(data)
            return_code = response == "Pong"
        except asyncio.TimeoutError:
            pass

        socket.close(linger=0)
    except Exception as exc:
        _LOGGER.warning(f"Caught an exception while pinging a control server at {endpoint}: {exc}.")

    return return_code


# Synchronous version for backward compatibility
def is_control_server_active_sync(endpoint: str = None, timeout: float = 0.5) -> bool:
    """Synchronous version of is_control_server_active for backward compatibility.

    This function runs the async version in a new event loop.
    """
    if endpoint is None:
        raise ValueError(
            "endpoint argument not provided, please provide a string with this format: '<protocol>://<address>:<port>'"
        )

    # Create a new event loop for this function call
    loop = asyncio.new_event_loop()
    try:
        return loop.run_until_complete(is_control_server_active(endpoint, timeout))
    finally:
        loop.close()


# Helper function to periodically run a coroutine
async def do_every_async(interval: float, coroutine: Callable[[], Coroutine]) -> None:
    """Run a coroutine every interval seconds"""
    while True:
        await coroutine()
        await asyncio.sleep(interval)


class ControlServer(abc.ABC):
    """Base class for all device control servers and for the Storage Manager and Configuration Manager.

    A Control Server reads commands from a ZeroMQ socket and executes these commands by calling the `execute()` method
    of the commanding protocol class.

    The subclass shall define the following:

    - Define the device protocol class -> `self.device_protocol`
    - Bind the command socket to the device protocol -> `self.dev_ctrl_cmd_sock`
    - Register the command socket in the poll set -> `self.poller`

    """

    def __init__(self):
        """Initialisation of a new Control Server."""

        from egse.monitoring import MonitoringProtocol
        from egse.services import ServiceProtocol

        self._process_status = ProcessStatus()
        self._metrics_task = None  # Will be created in serve()

        # The logger will be overwritten by the subclass, if not, we use this logger with the name of the subclass.
        # That will help us to identify which subclass did not overwrite the logger attribute.

        self.logger = logging.getLogger(get_full_classname(self))

        self.listeners = Listeners()
        self.scheduled_tasks = []

        # Queue for sequential operations that must preserve ordering
        self.sequential_queue = asyncio.Queue()

        self.interrupted = False
        self.mon_delay = 1000  # Delay between publish status information [ms]
        self.hk_delay = 1000  # Delay between saving housekeeping information [ms]
        self.scheduled_task_delay = 10  # delay time between successive executions of scheduled tasks [seconds]

        self.zcontext = zmq.asyncio.Context.instance()

        # No need for explicit poller in asyncio version
        # Instead, we'll use asyncio.gather with tasks for each socket

        self.device_protocol = None  # This will be set in the subclass
        self.service_protocol = ServiceProtocol(self)
        self.monitoring_protocol = MonitoringProtocol(self)

        # Set up the Control Server waiting for service requests
        self.dev_ctrl_service_sock = self.zcontext.socket(zmq.REP)
        self.service_protocol.bind(self.dev_ctrl_service_sock)

        # Set up the Control Server for sending monitoring info
        self.dev_ctrl_mon_sock = self.zcontext.socket(zmq.PUB)
        self.monitoring_protocol.bind(self.dev_ctrl_mon_sock)

        # Set up the Control Server waiting for device commands.
        # The device protocol shall bind the socket in the subclass
        self.dev_ctrl_cmd_sock = self.zcontext.socket(zmq.REP)

        # Tasks will be created in serve()
        self.tasks = []
        self.event_loop = None

    @abc.abstractmethod
    def get_communication_protocol(self) -> str:
        """Returns the communication protocol used by the Control Server.

        Returns:
            Communication protocol used by the Control Server, as specified in the settings.
        """

        pass

    @abc.abstractmethod
    def get_commanding_port(self) -> int:
        """Returns the commanding port used by the Control Server.

        Returns:
            Commanding port used by the Control Server, as specified in the settings.
        """

        pass

    @abc.abstractmethod
    def get_service_port(self) -> int:
        """Returns the service port used by the Control Server.

        Returns:
            Service port used by the Control Server, as specified in the settings.
        """

        pass

    @abc.abstractmethod
    def get_monitoring_port(self) -> int:
        """Returns the monitoring port used by the Control Server.

        Returns:
            Monitoring port used by the Control Server, as specified in the settings.
        """

        pass

    @staticmethod
    def get_ip_address() -> str:
        """Returns the IP address of the current host."""
        return get_host_ip()

    def get_storage_mnemonic(self) -> str:
        """Returns the storage mnemonics used by the Control Server.

        This is a string that will appear in the filename with the housekeeping information of the device, as a way of
        identifying the device.  If this is not implemented in the subclass, then the class name will be used.

        Returns:
            Storage mnemonics used by the Control Server, as specified in the settings.
        """

        return self.__class__.__name__

    def get_process_status(self) -> dict:
        """Returns the process status of the Control Server.

        Returns:
            Dictionary with the process status of the Control Server.
        """

        return self._process_status.as_dict()

    def get_average_execution_times(self) -> dict:
        """Returns the average execution times of all functions that have been monitored by this process.

        Returns:
            Dictionary with the average execution times of all functions that have been monitored by this process.
                The dictionary keys are the function names, and the values are the average execution times in ms.
        """

        return get_average_execution_times()

    def set_mon_delay(self, seconds: float) -> float:
        """Sets the delay time for monitoring.

        The delay time is the time between two successive executions of the `get_status()` function of the device
        protocol.

        It might happen that the delay time that is set is longer than what you requested. That is the case when the
        execution of the `get_status()` function takes longer than the requested delay time. That should prevent the
        server from blocking when a too short delay time is requested.

        Args:
            seconds (float): Number of seconds between the monitoring calls

        Returns:
            Delay that was set [ms].
        """

        execution_time = get_average_execution_time(self.device_protocol.get_status)
        self.mon_delay = max(seconds * 1000, (execution_time + 0.2) * 1000)

        return self.mon_delay

    def set_hk_delay(self, seconds: float) -> float:
        """Sets the delay time for housekeeping.

        The delay time is the time between two successive executions of the `get_housekeeping()` function of the device
        protocol.

        It might happen that the delay time that is set is longer than what you requested. That is the case when the
        execution of the `get_housekeeping()` function takes longer than the requested delay time. That should prevent
        the server from blocking when a too short delay time is requested.

        Args:
            seconds (float): Number of seconds between the housekeeping calls

        Returns:
            Delay that was set [ms].
        """

        execution_time = get_average_execution_time(self.device_protocol.get_housekeeping)
        self.hk_delay = max(seconds * 1000, (execution_time + 0.2) * 1000)

        return self.hk_delay

    def set_scheduled_task_delay(self, seconds: float):
        """
        Sets the delay time between successive executions of scheduled tasks.

        Args:
            seconds: the time interval between two successive executions [seconds]

        """
        self.scheduled_task_delay = seconds

    def set_logging_level(self, level: Union[int, str]) -> None:
        """Sets the logging level to the given level.

        Allowed logging levels are:

        - "CRITICAL" or "FATAL" or 50
        - "ERROR" or 40
        - "WARNING" or "WARN" or 30
        - "INFO" or 20
        - "DEBUG" or 10
        - "NOTSET" or 0

        Args:
            level (int | str): Logging level to use, specified as either a string or an integer
        """

        self.logger.setLevel(level=level)

    def quit(self) -> None:
        """Interrupts the Control Server."""

        self.interrupted = True
        if self.event_loop:
            for task in self.tasks:
                if not task.done():
                    task.cancel()

    async def before_serve(self) -> None:
        """
        This method needs to be overridden by the subclass if certain actions need to be executed before the control
        server is activated.
        """

        pass

    async def after_serve(self) -> None:
        """
        This method needs to be overridden by the subclass if certain actions need to be executed after the control
        server has been deactivated.
        """

        pass

    async def is_storage_manager_active(self) -> bool:
        """Checks if the Storage Manager is active.

        This method has to be implemented by the subclass if you need to store information.

        Note: You might want to set a specific timeout when checking for the Storage Manager.

        Note: If this method returns True, the following methods shall also be implemented by the subclass:

        - register_to_storage_manager()
        - unregister_from_storage_manager()
        - store_housekeeping_information()

        Returns:
            True if the Storage Manager is active; False otherwise.
        """

        return False

    async def handle_scheduled_tasks(self):
        """
        Executes or reschedules tasks in the `serve()` event loop.
        """
        self.scheduled_tasks.reverse()
        rescheduled_tasks = []
        while self.scheduled_tasks:
            task_info = self.scheduled_tasks.pop()
            task = task_info["task"]
            task_name = task_info.get("name")

            at = task_info.get("after")
            if at and at > datetime.datetime.now(tz=datetime.timezone.utc):
                rescheduled_tasks.append(task_info)
                continue

            condition = task_info.get("when")
            if condition and not condition():
                self.logger.debug(
                    f"Task {task_name} rescheduled in {self.scheduled_task_delay}s, condition not met...."
                )
                self.logger.info(f"Task {task_name} rescheduled in {self.scheduled_task_delay}s")
                current_time = datetime.datetime.now(tz=datetime.timezone.utc)
                scheduled_time = current_time + datetime.timedelta(seconds=self.scheduled_task_delay)
                task_info["after"] = scheduled_time
                rescheduled_tasks.append(task_info)
                continue

            self.logger.debug(f"Running scheduled task: {task_name}")
            try:
                # Handle both regular functions and coroutines
                if asyncio.iscoroutinefunction(task):
                    await task()
                else:
                    task()
            except Exception as exc:
                self.logger.error(f"Task {task_name} has failed: {exc!r}")
                self.logger.info(f"Task {task_name} rescheduled in {self.scheduled_task_delay}s")
                current_time = datetime.datetime.now(tz=datetime.timezone.utc)
                scheduled_time = current_time + datetime.timedelta(seconds=self.scheduled_task_delay)
                task_info["after"] = scheduled_time
                rescheduled_tasks.append(task_info)
            else:
                self.logger.debug(f"Scheduled task finished: {task_name}")

        if self.scheduled_tasks:
            self.logger.warning(f"There are still {len(self.scheduled_tasks)} scheduled tasks.")

        if rescheduled_tasks:
            self.scheduled_tasks.extend(rescheduled_tasks)

    def schedule_task(self, callback: Union[Callable, Coroutine], after: float = 0.0, when: Callable = None):
        """
        Schedules a task to run in the control server event loop.

        The `callback` function will be executed as soon as possible in the `serve()` event loop.

        Some simple scheduling options are available:

        * after: the task will only execute 'x' seconds after the time of scheduling. I.e.
          the task will be rescheduled until time > scheduled time + 'x' seconds.
        * when: the task will only execute when the condition is True.

        The `after` and the `when` arguments can be combined.

        Note:
            * This function is intended to be used in order to prevent a deadlock.
            * Since the `callback` function is executed in the `serve()` event loop, it shall not block!

        """
        try:
            name = callback.func.__name__ if isinstance(callback, partial) else callback.__name__
        except AttributeError:
            name = "unknown"

        current_time = datetime.datetime.now(tz=datetime.timezone.utc)
        scheduled_time = current_time + datetime.timedelta(seconds=after)

        self.logger.info(f"Task {name} scheduled")

        self.scheduled_tasks.append({"task": callback, "name": name, "after": scheduled_time, "when": when})

    async def process_device_command(self):
        """Handle commands for the device protocol"""
        while not self.interrupted:
            try:
                # Check if there's a command pending with non-blocking recv
                try:
                    # Use poll with a short timeout to check for messages
                    events = await self.dev_ctrl_cmd_sock.poll(timeout=50, flags=zmq.POLLIN)
                    if events == zmq.POLLIN:
                        # If we have a command, we can either:
                        # 1. Process it directly (parallel to other operations)
                        # await self.device_protocol.execute_async()

                        # 2. Or enqueue it for sequential processing if order matters
                        self.enqueue_sequential_operation(self.device_protocol.execute_async)
                except Exception as exc:
                    self.logger.error(f"Error checking for device command: {exc}")
                    await asyncio.sleep(0.05)

                await asyncio.sleep(0.01)  # Short sleep to prevent CPU hogging
            except asyncio.CancelledError:
                break
            except Exception as exc:
                self.logger.error(f"Error in device command processor: {exc}")

    async def process_service_command(self):
        """Handle commands for the service protocol"""
        while not self.interrupted:
            try:
                # Check if there's a command pending with non-blocking recv
                try:
                    # Use poll with a short timeout to check for messages
                    events = await self.dev_ctrl_service_sock.poll(timeout=50, flags=zmq.POLLIN)
                    if events == zmq.POLLIN:
                        # If we have a command, we can either:
                        # 1. Process it directly (parallel to other operations)
                        # await self.service_protocol.execute_async()

                        # 2. Or enqueue it for sequential processing if order matters
                        self.enqueue_sequential_operation(self.service_protocol.execute_async)
                except Exception as exc:
                    self.logger.error(f"Error checking for service command: {exc}")
                    await asyncio.sleep(0.05)

                await asyncio.sleep(0.01)  # Short sleep to prevent CPU hogging
            except asyncio.CancelledError:
                break
            except Exception as exc:
                self.logger.error(f"Error in service command processor: {exc}")

    async def send_status_updates(self):
        """Send periodic status updates"""
        while not self.interrupted:
            try:
                # Convert milliseconds to seconds for asyncio.sleep
                await asyncio.sleep(self.mon_delay / 1000)

                # Create a coroutine for the status update
                async def status_update_operation():
                    try:
                        status = save_average_execution_time(self.device_protocol.get_status)
                        await self.monitoring_protocol.send_status_async(status)
                    except Exception as exc:
                        _LOGGER.error(
                            textwrap.dedent(
                                f"""\
                                An Exception occurred while collecting status info from the control server \
                                 {self.__class__.__name__}.
                                This might be a temporary problem, still needs to be looked into:

                                {exc}
                                """
                            )
                        )

                # You can choose to run status updates sequentially if they must be in order
                # with other operations from the polling loop
                self.enqueue_sequential_operation(status_update_operation)

                # Or run them independently if order doesn't matter:
                # await status_update_operation()

            except asyncio.CancelledError:
                break
            except Exception as exc:
                self.logger.error(f"Error sending status: {exc}")

    async def send_housekeeping_updates(self):
        """Send periodic housekeeping updates if storage manager is active"""
        storage_manager = await self.is_storage_manager_active()
        if not storage_manager:
            return

        while not self.interrupted:
            try:
                # Convert milliseconds to seconds for asyncio.sleep
                await asyncio.sleep(self.hk_delay / 1000)

                # Create a coroutine for the housekeeping update
                async def housekeeping_update_operation():
                    try:
                        housekeeping = save_average_execution_time(self.device_protocol.get_housekeeping)
                        await self.store_housekeeping_information(housekeeping)
                    except Exception as exc:
                        _LOGGER.error(
                            textwrap.dedent(
                                f"""\
                                An Exception occurred while collecting housekeeping from the device to be stored in \
                                 {self.get_storage_mnemonic()}.
                                This might be a temporary problem, still needs to be looked into:

                                {exc}
                                """
                            )
                        )

                # You can choose to run housekeeping updates sequentially if they must be in order
                # with other operations from the polling loop
                self.enqueue_sequential_operation(housekeeping_update_operation)

                # Or run them independently if order doesn't matter:
                # await housekeeping_update_operation()

            except asyncio.CancelledError:
                break
            except Exception as exc:
                self.logger.error(f"Error sending housekeeping: {exc}")

    async def process_scheduled_tasks(self):
        """Process scheduled tasks periodically"""
        while not self.interrupted:
            try:
                # Create a coroutine for handling scheduled tasks
                async def scheduled_tasks_operation():
                    await self.handle_scheduled_tasks()

                # You can choose to run scheduled tasks sequentially if they must
                # maintain order with other operations from the polling loop
                self.enqueue_sequential_operation(scheduled_tasks_operation)

                # Or run them independently if order doesn't matter:
                # await scheduled_tasks_operation()

                await asyncio.sleep(0.05)  # Small sleep to not hog CPU
            except asyncio.CancelledError:
                break
            except Exception as exc:
                self.logger.error(f"Error processing scheduled tasks: {exc}")

    async def update_metrics(self):
        """Update process metrics periodically"""
        while not self.interrupted:
            try:
                self._process_status.update()
                await asyncio.sleep(PROCESS_SETTINGS.METRICS_INTERVAL)
            except asyncio.CancelledError:
                break
            except Exception as exc:
                self.logger.error(f"Error updating metrics: {exc}")

    async def check_device_protocol_alive(self):
        """Check if device protocol is still alive"""
        while not self.interrupted:
            try:
                await asyncio.sleep(1.0)  # Check every second

                if not self.device_protocol.is_alive():
                    self.logger.error(
                        "Some Thread or sub-process that was started by Protocol has died, terminating..."
                    )
                    self.quit()
                    break
            except asyncio.CancelledError:
                break
            except Exception as exc:
                self.logger.error(f"Error checking if device protocol is alive: {exc}")

    async def process_sequential_queue(self):
        """
        Process operations that need to be executed sequentially.
        This ensures that certain operations maintain the same order as in the original polling loop.
        """
        while not self.interrupted:
            try:
                # Get operation from queue with timeout to allow checking for interruption
                try:
                    operation = await asyncio.wait_for(self.sequential_queue.get(), 0.1)
                    await operation()
                    self.sequential_queue.task_done()
                except asyncio.TimeoutError:
                    continue
            except asyncio.CancelledError:
                break
            except Exception as exc:
                self.logger.error(f"Error processing sequential operation: {exc}")

    def enqueue_sequential_operation(self, coroutine_func):
        """
        Add an operation to the sequential queue.
        This ensures the operation will run in order with other sequential operations.

        Args:
            coroutine_func: A coroutine function (async function) to be executed sequentially
        """
        if self.sequential_queue is not None:  # Check if server is initialized
            self.sequential_queue.put_nowait(coroutine_func)

    async def serve(self) -> None:
        """Activation of the Control Server.

        This comprises the following steps:

        - Executing the `before_serve` method;
        - Checking if the Storage Manager is active and registering the Control Server to it;
        - Start accepting (listening to) commands;
        - Start sending out monitoring information;
        - Start sending out housekeeping information;
        - Start listening for quit commands;
        - After a quit command has been received:
            - Unregister from the Storage Manager;
            - Execute the `after_serve` method;
            - Close all sockets;
            - Clean up all tasks.
        """
        # Store reference to event loop
        self.event_loop = asyncio.get_event_loop()

        # Execute before_serve hook
        await self.before_serve()

        # Check if Storage Manager is available
        storage_manager = await self.is_storage_manager_active()
        if storage_manager:
            await self.register_to_storage_manager()

        # Set up signal handler
        killer = SignalCatcher()

        # Create tasks for each aspect of the control server
        self.tasks = [
            asyncio.create_task(self.process_device_command()),
            asyncio.create_task(self.process_service_command()),
            asyncio.create_task(self.send_status_updates()),
            asyncio.create_task(self.process_scheduled_tasks()),
            asyncio.create_task(self.update_metrics()),
            asyncio.create_task(self.check_device_protocol_alive()),
            asyncio.create_task(self.process_sequential_queue()),  # Add sequential queue processor
        ]

        # Add housekeeping task if storage manager is active
        if storage_manager:
            self.tasks.append(asyncio.create_task(self.send_housekeeping_updates()))

        # Wait for interruption or signal
        try:
            while not self.interrupted and not killer.term_signal_received:
                await asyncio.sleep(0.1)

                if killer.term_signal_received:
                    self.logger.info(f"TERM Signal received, closing down the {self.__class__.__name__}.")
                    break

                if self.interrupted:
                    self.logger.info(f"Quit command received, closing down the {self.__class__.__name__}.")
                    break

        except asyncio.CancelledError:
            self.logger.info("Main server loop cancelled.")
        finally:
            # Cancel all running tasks
            for task in self.tasks:
                if not task.done():
                    task.cancel()

            # Wait for tasks to complete their cancellation
            if self.tasks:
                try:
                    await asyncio.gather(*self.tasks, return_exceptions=True)
                except asyncio.CancelledError:
                    pass

            # Cleanup
            if storage_manager:
                await self.unregister_from_storage_manager()

            await self.after_serve()

            await self.device_protocol.quit_async()

            self.dev_ctrl_mon_sock.close(linger=0)
            self.dev_ctrl_service_sock.close(linger=0)
            self.dev_ctrl_cmd_sock.close(linger=0)

            close_all_zmq_handlers()

            self.zcontext.term()

    async def store_housekeeping_information(self, data: dict) -> None:
        """Sends housekeeping information to the Storage Manager.

        This method has to be overwritten by the subclasses if they want the device housekeeping information to be
        saved.

        Args:
            data (dict): a dictionary containing parameter name and value of all device housekeeping. There is also
                a timestamp that represents the date/time when the HK was received from the device.
        """
        pass

    async def register_to_storage_manager(self) -> None:
        """Registers this Control Server to the Storage Manager.

        By doing so, the housekeeping information of the device will be sent to the Storage Manager, which will store
        the information in a dedicated CSV file.

        This method has to be overwritten by the subclasses if they have housekeeping information that must be stored.

        Subclasses need to overwrite this method if they have housekeeping information to be stored.

        The following information is required for the registration:

        - origin: Storage mnemonic, which can be retrieved from `self.get_storage_mnemonic()`
        - persistence_class: Persistence layer (one of the TYPES in egse.storage.persistence)
        - prep: depending on the type of the persistence class (see respective documentation)

        The `egse.storage` module provides a convenience method that can be called from the method in the subclass:

            >>> from egse.storage import register_to_storage_manager_async  # noqa

        Note:
            the `egse.storage` module might not be available, it is provided by the `cgse-core` package.
        """
        pass

    async def unregister_from_storage_manager(self) -> None:
        """Unregisters the Control Server from the Storage Manager.

        This method has to be overwritten by the subclasses.

        The following information is required for the registration:

        - origin: Storage mnemonic, which can be retrieved from `self.get_storage_mnemonic()`

        The `egse.storage` module provides a convenience method that can be called from the method in the subclass:

            >>> from egse.storage import unregister_from_storage_manager_async  # noqa

        Note:
            the `egse.storage` module might not be available, it is provided by the `cgse-core` package.
        """
        pass

    async def notify_listeners(self, event_id: int = 0, context: dict = None):
        """
        Notifies registered listeners about an event.

        This function creates an Event object with the provided `event_id` and `context`
        and notifies all registered listeners with the created event.

        Args:
            event_id (int, optional): The identifier for the event. Defaults to 0.
            context (dict, optional): Additional context information associated with the event.
                Defaults to None.

        Note:
            The notification is performed by the `notify_listeners` method of the `listeners` object
            associated with this instance.
            The notification is executed as a task to avoid blocking the commanding chain.
        """
        self.logger.info(f"Notifying listeners for {EVENT_ID(event_id).name}")

        # Create a task for notification instead of a daemon thread
        event = Event(event_id=event_id, context=context)
        asyncio.create_task(self.listeners.notify_listeners_async(event))

    def get_listener_names(self):
        return self.listeners.get_listener_names()

    async def register_as_listener(self, proxy: Type, listener: dict):
        """
        Registers a listener with the specified proxy.

        This function attempts to add the provided listener to the specified proxy.
        It employs a retry mechanism to handle potential ConnectionError exceptions.

        Args:
            proxy: A callable object representing the proxy to which the listener will be added.
            listener (dict): The listener to be registered. Should be a dictionary containing
                listener details.

        Raises:
            ConnectionError: If the connection to the proxy encounters issues even after
                multiple retry attempts.
        """

        @retry_with_exponential_backoff(exceptions=[ConnectionError])
        async def _add_listener(proxy, listener):
            async with proxy() as x, x.get_service_proxy() as srv:
                rc = await srv.add_listener_async(listener)
                _LOGGER.info(f"Response from {proxy.__name__} service add_listener: {rc}")

        _LOGGER.info(f"Registering {self.__class__.__name__} as a listener to {proxy.__name__}")
        await _add_listener(proxy, listener)

    async def unregister_as_listener(self, proxy: Type, listener: dict):
        """
        Removes a registered listener from the specified proxy.

        This function attempts to remove the provided listener from the specified proxy.
        It employs a retry mechanism to handle potential ConnectionError exceptions.

        Args:
            proxy: A callable object representing the proxy from which the listener will be removed.
            listener (dict): The listener to be removed. Should be a dictionary containing
                listener details.

        Raises:
            ConnectionError: If the connection to the proxy encounters issues even after
                multiple retry attempts.
        """

        @retry(times=5, exceptions=[ConnectionError])
        async def _remove_listener(proxy, listener):
            async with proxy() as x, x.get_service_proxy() as srv:
                rc = await srv.remove_listener_async(listener)
                _LOGGER.debug(f"Response from remove_listener: {rc=}")

        # Check if the CS is available
        try:
            async with proxy():
                pass
        except ConnectionError:
            _LOGGER.warning(
                f"The {proxy.__class__.__name__} endpoint is not responding, {listener['name']} not un-registered."
            )
            return

        _LOGGER.info(f"Removing {self.__class__.__name__} as a listener from {proxy.__name__}")
        await _remove_listener(proxy, listener)



import subprocess

import sys
from pathlib import Path
from typing import Annotated

import rich
import typer

from egse.env import get_log_file_location

lakeshore336 = typer.Typer(
    name="LakeShore336",
    help="LakeShore336 Data Acquisition Unit, LakeShore, temperature monitoring",
)


@lakeshore336.command(name="start")
def start_lakeshore336(
    device_id: Annotated[str, typer.Argument(help="the device identifier, identifies the hardware controller")],
    simulator: Annotated[
        bool, typer.Option("--simulator", "--sim", help="Start the LakeShore336 simulator as the backend")
    ] = False,
    background: Annotated[bool, typer.Option(help="Start the LakeShore336 in the background")] = False,
):
    """Starts the LakeShore336 Control Server.

    This Control Server is started in the background by default.

    Args:
        device_id (str): Device identifier
        simulator (bool): Indicates whether the LakeShore336 Control Server should be started in simulator mode
        background (bool): Indicates whether the LakeShore336 Control Server should be started in the background
    """

    rich.print(f"Starting service LakeShore336 {device_id}")

    if background:
        location = get_log_file_location()
        output_filename = "lakeshore336_cs.start.out"
        output_path = Path(location, output_filename).expanduser()

        rich.print(f"Starting the LakeShore336 Control Server ({device_id}) – {simulator = }")
        rich.print(f"Output will be redirected to {output_path!s}")

        out = open(output_path, "w")

        cmd = [sys.executable, "-m", "egse.tempcontrol.lakeshore.lakeshore336_cs", "start", device_id]
        if simulator:
            cmd.append("--simulator")

        subprocess.Popen(cmd, stdout=out, stderr=out, stdin=subprocess.DEVNULL, close_fds=True)

    else:
        from egse.tempcontrol.lakeshore import lakeshore336_cs

        lakeshore336_cs.start(device_id, simulator)


@lakeshore336.command(name="stop")
def stop_lakeshore336(
    device_id: Annotated[str, typer.Argument(help="the device identifier, identifies the hardware controller")],
):
    """Stops the LakeShore336 service.

    Args:
        device_id (str): Device identifier
    """

    rich.print(f"Terminating service LakeShore336 {device_id}")

    from egse.tempcontrol.lakeshore import lakeshore336_cs

    lakeshore336_cs.stop(device_id)


@lakeshore336.command(name="status")
def status_lakeshore336(
    device_id: Annotated[str, typer.Argument(help="the device identifier, identifies the hardware controller")],
):
    """Prints status information on the LakeShore336 service.

    Args:
        device_id (str): Device identifier
    """

    rich.print(f"Status of LakeShore336 {device_id}")

    from egse.tempcontrol.lakeshore import lakeshore336_cs

    lakeshore336_cs.status(device_id)


@lakeshore336.command(name="start-simulator")
def start_lakeshore336_sim(
    device_id: Annotated[str, typer.Argument(help="the device identifier, identifies the hardware controller")],
):
    """Starts the LakeShore336 Simulator.

    Args:
        device_id (str): Device identifier
    """

    rich.print("Starting service LakeShore336 Simulator")

    from egse.tempcontrol.lakeshore import lakeshore336_cs

    lakeshore336_cs.status(device_id)



import asyncio
import contextlib
import sys
import threading
import time

import pytest

from egse.log import logger
from egse.notifyhub import async_is_notify_hub_active
from egse.notifyhub import is_notify_hub_active
from egse.notifyhub.event import NotificationEvent
from egse.notifyhub.server import AsyncNotificationHub
from egse.notifyhub.services import AsyncEventPublisher
from egse.notifyhub.services import AsyncEventSubscriber
from egse.notifyhub.services import EventPublisher
from egse.notifyhub.services import EventSubscriber
from egse.process import SubProcess
from egse.registry import is_service_registry_active
from egse.system import Timer
from egse.system import type_name
from egse.system import waiting_for


pytestmark = pytest.mark.skipif(
    is_notify_hub_active(), reason="The notification hub shall NOT be running for this test."
)


@contextlib.asynccontextmanager
async def async_notify_hub():
    """Asynchronous context manager that starts a notification hub as an asyncio Task."""

    if await async_is_notify_hub_active():
        pytest.xfail("The notification hub shall not be running for this test.")

    server = AsyncNotificationHub()
    server_task = asyncio.create_task(server.start())

    with Timer(name="Notify Hub startup timer"):
        await asyncio.wait_for(async_is_notify_hub_active(), timeout=5.0)

    try:
        yield
    except Exception as exc:
        logger.error(f"Caught {type_name(exc)}: {exc}", exc_info=True)

    await server.stop()

    await asyncio.gather(server_task, return_exceptions=True)

    if not server_task.done():
        server_task.cancel()

    with contextlib.suppress(asyncio.CancelledError):
        await server_task


@contextlib.contextmanager
def notify_hub():
    """Context manager that starts a notification hub as a sub-process."""

    if is_notify_hub_active():
        pytest.xfail("The notification hub shall not be running for this test.")

    proc = SubProcess("Notification Hub", [sys.executable, "-m", "egse.notifyhub.server"], ["start"])
    proc.execute()

    with Timer(name="Notify Hub startup timer"):
        waiting_for(is_notify_hub_active, timeout=5.0)

    yield proc

    proc.quit()

    with Timer(name="Notify Hub shutdown timer"):
        waiting_for(lambda: not is_notify_hub_active(), timeout=5.0)


def single_event_handler(event_data: dict):
    logger.info(f"{event_data=}")
    assert event_data["event_type"] == "single-event"
    assert event_data["source_service"] == "test_notify_hub"
    assert event_data["data"]["data"] == "Simple string for a single event"


@pytest.mark.skipif(not is_service_registry_active(), reason="This test needs the service registry running")
@pytest.mark.asyncio
async def test_server_running_async():
    # This test starts the notification hub and let it running for 30s
    # In these 30s it should:
    #
    # - register as a service to the service registry
    # - send at least three heartbeats to the registry server
    # - de-register from the service registry

    async with async_notify_hub():
        logger.warning("Please note this test takes 30s, check the log afterwards for expected server actions.")
        await asyncio.sleep(30.0)


def test_single_event():
    with notify_hub():
        publisher = EventPublisher()
        publisher.connect()

        event = NotificationEvent(
            event_type="single-event",
            source_service="test_notify_hub",
            data={"data": "Simple string for a single event"},
        )

        def poll_next_event():
            subscriber = EventSubscriber(["single-event"])
            subscriber.connect()
            subscriber.register_handler("single-event", single_event_handler)
            event_received = False
            while not event_received:
                if subscriber.poll():
                    subscriber.handle_event()
                    event_received = True
                    continue

                logger.info("no event received yet...")
                time.sleep(1.0)

            subscriber.disconnect()

        thread = threading.Thread(target=poll_next_event)
        thread.start()

        time.sleep(0.1)  # give the thread time to start

        logger.info(f"Publishing event {event.event_type}")
        publisher.publish(event)

        time.sleep(1.0)

        publisher.disconnect()


@pytest.mark.asyncio
async def test_single_event_async():
    async with async_notify_hub():
        publisher = AsyncEventPublisher()
        await publisher.connect()

        subscriber = AsyncEventSubscriber(["single-event"])
        await subscriber.connect()

        subscriber.register_handler("single-event", single_event_handler)

        event_listener = asyncio.create_task(subscriber.start_listening())

        await asyncio.sleep(0.1)

        event = NotificationEvent(
            event_type="single-event",
            source_service="test_notify_hub",
            data={"data": "Simple string for a single event"},
        )
        await publisher.publish(event)

        await asyncio.sleep(0.2)

        subscriber.disconnect()

        await event_listener

        publisher.disconnect()



import numpy as np

from egse.coordinates import dict_to_ref_model
from egse.coordinates import deserialize_array
from egse.coordinates import serialize_array
from egse.coordinates import ref_model_to_dict
from egse.coordinates.referenceFrame import ReferenceFrame


def test_serialization_of_numpy_arrays():
    assert serialize_array([1, 2, 3]) == "[1, 2, 3]"
    assert serialize_array(np.array([1, 2, 3])) == "[1, 2, 3]"
    assert serialize_array([[1], [2], [3]]) == "[[1], [2], [3]]"
    assert serialize_array([[1, 2.3, 4], [5, 6.2, 7]]) == ("[[1.0000, 2.3000, 4.0000], [5.0000, 6.2000, 7.0000]]")


def test_deserialization_of_numpy_array():
    assert (deserialize_array("[1, 2, 3]") == np.array([1, 2, 3])).all()
    assert (deserialize_array("[[1], [2], [3]]") == np.array([[1], [2], [3]])).all()


def test_serialization_of_reference_model():
    rot_config = "sxyz"

    master = ReferenceFrame.createMaster()

    a_ref = ReferenceFrame(transformation=np.identity(4), ref=master, name="a_ref", rot_config=rot_config)
    a_ref.addLink(master)

    b_ref = ReferenceFrame(transformation=np.identity(4), ref=a_ref, name="b_ref", rot_config=rot_config)

    c_ref = ReferenceFrame.fromTranslationRotation(
        [-2, -2, -2], [-3, -4, -5], rot_config=rot_config, ref=b_ref, name="c_ref"
    )
    c_ref.addLink(b_ref)

    model_list = [a_ref, b_ref, c_ref, master]

    model_def = ref_model_to_dict(model_list)

    print()
    print(model_def.pretty_str())

    ref_model = dict_to_ref_model(model_def)

    assert "a_ref" in ref_model.keys()
    assert ref_model.a_ref.name == "a_ref"

    assert "b_ref" in ref_model.keys()
    assert "c_ref" in ref_model.keys()
    assert "Master" in ref_model.keys()

    print()
    print(ref_model.pretty_str())



import asyncio
import logging
import sys
import time
from typing import Annotated

import rich
import typer

from egse.decorators import async_timer
from egse.decorators import timer
from egse.registry.client import AsyncRegistryClient
from egse.registry.client import RegistryClient

logging.basicConfig(level=logging.INFO)


@timer()
def main(count: int = 1):
    for _ in range(count):
        with RegistryClient() as reg:
            if reg.health_check():
                rich.print("[green]registry server is healthy.")
            else:
                rich.print("[red]error: health check failed.[/]")

            service_id = reg.register("stress-test", "localhost", 1234, "testing")
            if service_id is None:
                rich.print("[red]error: Couldn't register stress-test[/]")

            endpoint = reg.get_endpoint("testing")
            if endpoint is None:
                rich.print("[red]error: no endpoint for testing[/]")

            service = reg.get_service(service_id)
            if service is None:
                rich.print(f"[red]error: no service {service_id} found[/]")

            if not reg.deregister(service_id):
                rich.print("[red]error: Couldn't de-register stress-test[/]")


@async_timer()
async def amain(count: int = 1):
    for _ in range(count):
        with AsyncRegistryClient() as reg:
            if await reg.health_check():
                rich.print("[green]registry server is healthy.")
            else:
                rich.print("[red]error: health check failed.[/]")

            service_id = await reg.register("stress-test", "localhost", 1234, "testing")
            if service_id is None:
                rich.print("[red]error: Couldn't register stress-test[/]")

            endpoint = await reg.get_endpoint("testing")
            if endpoint is None:
                rich.print("[red]error: no endpoint for testing[/]")

            service = await reg.get_service(service_id)
            if service is None:
                rich.print(f"[red]error: no service {service_id} found[/]")

            if not await reg.deregister(service_id):
                rich.print("[red]error: Couldn't de-register stress-test[/]")


app = typer.Typer()


@app.command()
def run_stress_test(
    count: Annotated[int, typer.Option(help="Number of times to execute the tests")] = 1,
    sleep_time: Annotated[float, typer.Option(help="sleep time between tests")] = 10.0,
    terminate_server: Annotated[
        bool, typer.Option(help="Terminate the registry server at the end of the test")
    ] = False,
):
    with RegistryClient() as reg:
        service_id = reg.register("stress-test-heartbeat", "localhost", 5678)
        reg.start_heartbeat(3)

        main(count)  # executes the synchronous registry client

        # Allow time for the heartbeats to be sent by this test script

        time.sleep(sleep_time)

        # This should show the stress-test-heartbeat in the registered services
        response = reg.server_status()
        rich.print(response)

        time.sleep(sleep_time)

        asyncio.run(amain(count))  # executes the asynchronous registry client

        reg.stop_heartbeat()
        reg.deregister(service_id)

        # We should not see the stress-test-heartbeat anymore in the registered services
        response = reg.server_status()
        rich.print(response)

        if terminate_server:
            if reg.terminate_registry_server():
                rich.print("[green]Registry Server terminated successfully[/]")
            else:
                rich.print("[red]error: Couldn't terminate registry server..[/]")


if __name__ == "__main__":
    sys.exit(app())
